"""
GatingStrategy class
"""
import json
import anytree
import pandas as pd
from anytree.exporter import DotExporter
import numpy as np
import networkx as nx
from .._models import dimension
# noinspection PyProtectedMember
from .._models.gates._base_gate import Gate
from .._models import gates as fk_gates
# noinspection PyProtectedMember
from .._models.transforms._base_transform import Transform
# noinspection PyProtectedMember
from .._models.transforms._matrix import Matrix
from .._models.gating_results import GatingResults


class GatingStrategy(object):
    """
    Represents a flow cytometry gating strategy, including instructions
    for compensation and transformation.
    """
    def __init__(self):
        self._cached_preprocessed_events = {}

        # keys are the object's ID (xform or matrix),
        # values are the object itself
        self.transformations = {}
        self.comp_matrices = {}

        self._gate_tree = anytree.Node('root')

        # use a directed acyclic graph for later processing and enforcing there
        # are no cyclical gate relationships.
        # Each node is made from a tuple representing it's full gate path
        self._dag = nx.DiGraph()
        self._dag.add_node(('root',))

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'{len(self._gate_tree.descendants)} gates, {len(self.transformations)} transforms, '
            f'{len(self.comp_matrices)} compensations)'
        )

    def add_gate(self, gate, gate_path=None):
        """
        Add a gate to the gating strategy, see `gates` module. The gate ID and gate path must be
        unique in the gating strategy. A gate with a unique gate ID and parent can be added without
        specifying a gate_path. However, if the gate's ID and parent combination already exists in
        the gating strategy, a unique gate path must be provided.

        :param gate: instance from a sub-class of the Gate class
        :param gate_path: complete tuple of gate IDs for unique set of gate ancestors.
            Required if gate.gate_name and gate.parent combination is ambiguous

        :return: None
        """
        if not isinstance(gate, Gate):
            raise ValueError("gate must be a sub-class of the Gate class")

        parent_gate_name = gate.parent
        if parent_gate_name is None:
            # If no parent gate is specified, use root
            parent_gate_name = 'root'

        # Verify the gate parent matches the last item in the gate path (if given)
        if gate_path is not None:
            if len(gate_path) != 0:
                if parent_gate_name != gate_path[-1]:
                    raise ValueError("The gate parent and the last item in gate path are different.")

        # Find simple case of matching gate name + parent where no gate_path is specified.
        matched_nodes = anytree.findall(
            self._gate_tree,
            filter_=lambda g_node:
            g_node.name == gate.gate_name and
            g_node.parent.name == parent_gate_name
        )
        match_count = len(matched_nodes)

        if match_count != 0 and gate_path is None:
            raise KeyError(
                "A gate with name '%s' and parent gate name '%s' is already defined. " 
                "You must specify a gate_path as a unique tuple of ancestors." % (gate.gate_name, parent_gate_name)
            )

        # Here we either have a unique gate ID + parent, or an ambiguous gate ID + parent with a gate path.
        # It is still possible that the given gate_path already exists, we'll check that.
        # We'll find the parent node from the ID, and if there are multiple parent matches then resort
        # to using the given gate path.
        if parent_gate_name == 'root':
            # Easiest case since a root parent is also the full path.
            parent_node = self._gate_tree
        else:
            # Find all nodes with the parent ID name. If there's only one, we've identified the correct parent.
            # If there are none, the parent doesn't exist (or is a Quad gate).
            # If there are >1, we have to compare the gate paths.
            matching_parent_nodes = anytree.search.findall_by_attr(self._gate_tree, parent_gate_name)
            matched_parent_count = len(matching_parent_nodes)
            match_idx = None

            if matched_parent_count == 0:
                # TODO: need to double-check whether this scenario could be in a quadrant gate
                raise ValueError("Parent gate %s does not exist in the gating strategy" % parent_gate_name)
            elif matched_parent_count == 1:
                # There's only one match for the parent, so we're done
                match_idx = 0
            elif matched_parent_count > 1:
                for i, matched_parent_node in enumerate(matching_parent_nodes):
                    matched_parent_ancestors = tuple((pn.name for pn in matched_parent_node.path))
                    if matched_parent_ancestors == gate_path:
                        match_idx = i
                        break

            # look up the parent node, then do one final check to make sure the new gate doesn't
            # already exist as a child of the parent
            parent_node = matching_parent_nodes[match_idx]
            parent_child_nodes = anytree.search.findall_by_attr(parent_node, gate.gate_name, maxlevel=1)
            if len(parent_child_nodes) > 0:
                raise ValueError(
                    "A gate already exists matching gate name %s and the specified gate path" % gate.gate_name
                )

        node = anytree.Node(gate.gate_name, parent=parent_node, gate=gate)
        parent_node_tuple = tuple(n.name for n in parent_node.path)
        new_node_tuple = parent_node_tuple + (gate.gate_name,)
        self._dag.add_node(new_node_tuple)
        self._dag.add_edge(parent_node_tuple, new_node_tuple)

        # Quadrant gates need special handling to add their individual quadrants as children.
        # Other gates cannot have the main quadrant gate as a parent, they can only reference
        # the individual quadrants as parents.
        if isinstance(gate, fk_gates.QuadrantGate):
            for q_id, q in gate.quadrants.items():
                anytree.Node(q_id, parent=node, gate=q)

                q_node_tuple = new_node_tuple + (q_id,)

                self._dag.add_node(q_node_tuple)
                self._dag.add_edge(new_node_tuple, q_node_tuple)

        if isinstance(gate, fk_gates.BooleanGate):
            bool_gate_refs = gate.gate_refs
            for gate_ref in bool_gate_refs:
                gate_ref_node_tuple = tuple(gate_ref['path']) + (gate_ref['ref'],)
                self._dag.add_edge(gate_ref_node_tuple, new_node_tuple)

    def add_transform(self, transform):
        """
        Add a transform to the gating strategy, see `transforms` module. The transform ID must be unique in the
        gating strategy.

        :param transform: instance from a sub-class of the Transform class
        :return: None
        """
        if not isinstance(transform, Transform):
            raise ValueError("transform must be a sub-class of the Transform class")

        if transform.id in self.transformations:
            raise KeyError("Transform ID '%s' is already defined" % transform.id)

        self.transformations[transform.id] = transform

    def get_transform(self, transform_id):
        """
        Retrieve transform instance from gating strategy.

        :param transform_id: String of a valid transform ID stored in the gating strategy
        :return: Instance of a Transform sub-class
        """
        return self.transformations[transform_id]

    def add_comp_matrix(self, matrix):
        """
        Add a compensation matrix to the gating strategy, see `transforms` module. The matrix ID must be unique in the
        gating strategy.

        :param matrix: an instance of the Matrix class
        :return: None
        """
        # Only accept Matrix class instances as we need the ID
        if not isinstance(matrix, Matrix):
            raise ValueError("matrix must be an instance of the Matrix class")

        if matrix.id in self.comp_matrices:
            raise KeyError("Matrix ID '%s' is already defined" % matrix.id)

        self.comp_matrices[matrix.id] = matrix

    def get_comp_matrix(self, matrix_id):
        """
        Retrieve Matrix instance from gating strategy.

        :param matrix_id: String of a valid Matrix ID stored in the gating strategy
        :return: Matrix instance
        """
        return self.comp_matrices[matrix_id]

    def _get_gate_node(self, gate_name, gate_path=None):
        node_matches = anytree.findall_by_attr(self._gate_tree, gate_name)
        node_match_count = len(node_matches)

        if node_match_count == 1:
            node = node_matches[0]
        elif node_match_count > 1:
            # need to match on full gate path
            # TODO: what if QuadrantGate is re-used, does this still work for that case
            if gate_path is None:
                raise ValueError(
                    "Found multiple gates with name %s. Provide full 'gate_path' to disambiguate." % gate_name
                )

            gate_matches = []
            gate_path_length = len(gate_path)
            for n in node_matches:
                if len(n.ancestors) != gate_path_length:
                    continue
                ancestor_matches = tuple((a.name for a in n.ancestors if a.name in gate_path))
                if ancestor_matches == gate_path:
                    gate_matches.append(n)

            if len(gate_matches) == 1:
                node = gate_matches[0]
            elif len(gate_matches) > 1:
                raise ValueError("Report as bug: Found multiple gates with ID %s and given gate path." % gate_name)
            else:
                node = None
        else:
            node = None

        if node is None:
            raise ValueError("Gate name %s was not found in gating strategy" % gate_name)

        return node

    def get_root_gates(self):
        """
        Retrieve list of root-level gate instances.

        :return: list of Gate instances
        """
        root = self._gate_tree.root
        root_children = root.children

        root_gates = []

        for node in root_children:
            root_gates.append(node.gate)

        return root_gates

    def get_gate(self, gate_name, gate_path=None):
        """
        Retrieve a gate instance by its gate ID.

        :param gate_name: text string of a gate name
        :param gate_path: complete tuple of gate IDs for unique set of gate ancestors.
            Required if gate_name is ambiguous
        :return: Subclass of a Gate object
        :raises KeyError: if gate ID is not found in gating strategy
        """
        node = self._get_gate_node(gate_name, gate_path)

        if isinstance(node.gate, fk_gates.Quadrant):
            # return the full QuadrantGate b/c a Quadrant by itself has no parent reference
            node = node.parent

        return node.gate

    def get_parent_gate(self, gate_name, gate_path=None):
        """
        Retrieve the parent Gate instance for the given gate ID.

        :param gate_name: text string of a gate name
        :param gate_path: complete tuple of gate IDs for unique set of gate ancestors.
            Required if gate_name is ambiguous
        :return: Subclassed Gate instance
        """
        node = self._get_gate_node(gate_name, gate_path)

        if node.parent.name == 'root':
            return None

        return node.parent.gate

    def get_child_gates(self, gate_name, gate_path=None):
        """
        Retrieve list of child gate instances by their parent's gate ID.

        :param gate_name: text string of a gate name
        :param gate_path: complete tuple of gate IDs for unique set of gate ancestors.
            Required if gate_name is ambiguous
        :return: list of Gate instances
        :raises KeyError: if gate ID is not found in gating strategy
        """
        node = self._get_gate_node(gate_name, gate_path)

        child_gates = []

        for n in node.children:
            child_gates.append(n.gate)

        return child_gates

    def get_gate_ids(self):
        """
        Retrieve the list of gate IDs (with ancestors) for the gating strategy
        :return: list of tuples where the 1st item is the gate name string and 2nd item is
                 a list of ancestor gate names
        """
        gates = []
        for node in self._gate_tree.descendants:
            ancestors = tuple((a.name for a in node.ancestors))
            gates.append((node.name, ancestors))

        return gates

    def get_gate_hierarchy(self, output='ascii', **kwargs):
        """
        Retrieve the hierarchy of gates in the gating strategy in several formats, including text,
        dictionary, or JSON. If output == 'json', extra keyword arguments are passed to json.dumps

        :param output: Determines format of hierarchy returned, either 'ascii',
            'dict', or 'JSON' (default is 'ascii')
        :return: gate hierarchy as a text string or a dictionary
        """
        if output == 'ascii':
            tree = anytree.RenderTree(self._gate_tree, style=anytree.render.ContRoundStyle())
            lines = []

            for row in tree:
                lines.append("%s%s" % (row.pre, row.node.name))

            return "\n".join(lines)
        elif output == 'json':
            dict_exporter = anytree.exporter.DictExporter(
                attriter=lambda attrs: [(k, v) for k, v in attrs if k != 'gate']
            )

            gs_dict = dict_exporter.export(self._gate_tree)
            gs_json = json.dumps(gs_dict, **kwargs)

            return gs_json
        elif output == 'dict':
            exporter = anytree.exporter.DictExporter()
            gs_dict = exporter.export(self._gate_tree)

            return gs_dict
        else:
            raise ValueError("'output' must be either 'ascii', 'json', or 'dict'")

    def export_gate_hierarchy_image(self, output_file_path):
        """
        Saves an image of the gate hierarchy in many common formats
        according to the extension given in `output_file_path`, including

            - SVG  ('svg')
            - PNG  ('png')
            - JPEG ('jpeg', 'jpg')
            - TIFF ('tiff', 'tif')
            - GIF  ('gif')
            - PS   ('ps')
            - PDF  ('pdf')

        *Requires that `graphviz` is installed.*

        :param output_file_path: File path (including file name) of image
        :return: None
        """
        DotExporter(self._gate_tree).to_picture(output_file_path)

    def _get_cached_preprocessed_events(self, sample_id, comp_ref, xform_ref, dim_idx=None):
        """
        Retrieve cached pre-processed events (if they exist)

        :param sample_id: a text string for a Sample ID
        :param comp_ref: text string for a Matrix ID
        :param xform_ref: text string for a Transform ID
        :param dim_idx: channel index for requested preprocessed events. If None, all channels are requested.
        :return: NumPy array of the cached compensated events for the given sample, None if no cache exists.
        """
        ref_tuple = (comp_ref, xform_ref, dim_idx)

        try:
            # return a copy of cached events in case downstream modifies them
            return self._cached_preprocessed_events[sample_id][ref_tuple].copy()
        except KeyError:
            return None

    def clear_cache(self):
        """
        Clears all cached pre-processed events stored in the GatingStrategy. This is useful to
        reduce memory usage after analyzing large data sets. Clearing the cache will not affect
        any results previously retrieved.
        :return: None
        """
        self._cached_preprocessed_events = {}

    def _cache_preprocessed_events(self, preprocessed_events, sample_id, comp_ref, xform_ref, dim_idx=None):
        """
        Cache pre-processed events for a Sample instance

        :param preprocessed_events: NumPy array of the cached compensated events for the given sample
        :param sample_id: a text string for a Sample ID
        :param comp_ref: text string for a Matrix ID
        :param xform_ref: text string for a Transform ID
        :param dim_idx: channel index for given preprocessed events. If None, all channels where preprocessed
        :return: None
        """
        if sample_id not in self._cached_preprocessed_events:
            self._cached_preprocessed_events[sample_id] = {}

        sample_cache = self._cached_preprocessed_events[sample_id]
        ref_tuple = (comp_ref, xform_ref, dim_idx)

        if ref_tuple not in sample_cache:
            # copy the given events here to decouple from any use analysis
            sample_cache[ref_tuple] = preprocessed_events.copy()

    def _compensate_sample(self, dim_comp_refs, sample):
        dim_comp_ref_count = len(dim_comp_refs)

        if dim_comp_ref_count == 0:
            events = sample.get_events(source='raw')
            return events.copy()
        elif dim_comp_ref_count > 1:
            raise NotImplementedError(
                "Mixed compensation between individual channels is not "
                "implemented. Never seen it, but if you are reading this "
                "message, submit an issue to have it implemented."
            )
        else:
            comp_ref = list(dim_comp_refs)[0]

        # noinspection PyProtectedMember
        events = self._get_cached_preprocessed_events(
            sample.original_filename,
            comp_ref,
            None,
            None
        )

        if events is not None:
            return events

        if comp_ref == 'FCS':
            meta = sample.get_metadata()

            if 'spill' not in meta or 'spillover' not in meta:
                # GML 2.0 spec states if 'FCS' is specified but no spill is present, treat as uncompensated
                events = sample.get_events(source='raw')
                return events.copy()

            try:
                spill = meta['spillover']  # preferred, per FCS standard
            except KeyError:
                spill = meta['spill']

            detectors = [sample.pnn_labels[i] for i in sample.fluoro_indices]
            fluorochromes = [sample.pns_labels[i] for i in sample.fluoro_indices]
            matrix = Matrix('fcs', spill, detectors, fluorochromes, null_channels=sample.null_channels)
        else:
            # lookup specified comp-ref in gating strategy
            matrix = self.comp_matrices[comp_ref]

        if matrix is not None:
            events = matrix.apply(sample)
            # cache the comp events
            # noinspection PyProtectedMember
            self._cache_preprocessed_events(
                events,
                sample.original_filename,
                comp_ref,
                None,
                None
            )

        return events

    def _preprocess_sample_events(self, sample, gate, cache_events=False):
        pnn_labels = sample.pnn_labels
        pns_labels = sample.pns_labels
        # FlowJo replaces slashes with underscores, so make a set of labels with that replacement
        flowjo_pnn_labels = [label.replace('/', '_') for label in pnn_labels]

        dim_indices = []
        dim_ids = []
        dim_comp_refs = set()
        new_dims = []
        new_dim_ids = []
        dim_xform = []

        for dim in gate.dimensions:
            dim_comp = False
            if dim.compensation_ref not in [None, 'uncompensated']:
                dim_comp_refs.add(dim.compensation_ref)
                dim_comp = True

            if isinstance(dim, dimension.RatioDimension):
                # dimension is a transform of other dimensions
                new_dims.append(dim)
                new_dim_ids.append(dim.ratio_ref)
                continue
            elif isinstance(dim, dimension.QuadrantDivider):
                dim_id = dim.dimension_ref
            else:
                dim_id = dim.id

            if dim_id in pnn_labels:
                dim_indices.append(pnn_labels.index(dim_id))
            elif dim_id in pns_labels:
                dim_indices.append(pns_labels.index(dim_id))
            elif dim_id in flowjo_pnn_labels:
                dim_indices.append(flowjo_pnn_labels.index(dim_id))
            else:
                # for a referenced comp, the label may have been the
                # fluorochrome instead of the channel's PnN label. If so,
                # the referenced matrix object will also have the detector
                # names that will match
                if not dim_comp:
                    raise LookupError(
                        "%s is not found as a channel label or channel reference in %s" % (dim_id, sample)
                    )
                matrix = self.comp_matrices[dim.compensation_ref]
                try:
                    matrix_dim_idx = matrix.fluorochomes.index(dim_id)
                except ValueError:
                    raise ValueError("%s not found in list of matrix fluorochromes" % dim_id)
                detector = matrix.detectors[matrix_dim_idx]
                dim_indices.append(pnn_labels.index(detector))

            dim_ids.append(dim_id)

            dim_xform.append(dim.transformation_ref)

        # compensate will cache events regardless of cache_events arg value
        # this is much more universally useful to speed up analysis as the
        # same comp matrix is nearly always used for all sample gates.
        events = self._compensate_sample(dim_comp_refs, sample)

        for i, dim in enumerate(dim_indices):
            if dim_xform[i] is not None:
                if len(dim_comp_refs) > 0:
                    comp_ref = list(dim_comp_refs)[0]
                else:
                    comp_ref = None

                xform = self.transformations[dim_xform[i]]

                if cache_events:
                    cached_events = self._get_cached_preprocessed_events(
                        sample.original_filename,
                        comp_ref,
                        dim_xform[i],
                        dim_idx=dim
                    )
                else:
                    cached_events = None

                if cached_events is not None:
                    events[:, [dim]] = cached_events
                else:
                    xform_events = xform.apply(events[:, [dim]])
                    events[:, [dim]] = xform_events
                    if cache_events:
                        self._cache_preprocessed_events(
                            xform_events,
                            sample.original_filename,
                            comp_ref,
                            dim_xform[i],
                            dim_idx=dim
                        )

        df_events = pd.DataFrame(events[:, dim_indices], columns=dim_ids)
        if len(new_dims) > 0:
            new_dim_events = self._process_new_dims(sample, new_dims)
            df_new_dim_events = pd.DataFrame(new_dim_events, columns=new_dim_ids)
            df_events = pd.concat([df_events, df_new_dim_events], axis=1)

        return df_events

    def _process_new_dims(self, sample, new_dims):
        new_dims_events = []

        for new_dim in new_dims:
            # Allows gate classes to handle new dimensions created from
            # ratio transforms. Note, the ratio transform's apply method is
            # different from other transforms in that it takes a sample argument
            # and not an events argument

            # new dimensions are defined by transformations of other dims
            try:
                new_dim_xform = self.transformations[new_dim.ratio_ref]
            except KeyError:
                raise KeyError("New dimensions must provide a transformation")

            xform_events = new_dim_xform.apply(sample)

            if new_dim.transformation_ref is not None:
                xform = self.transformations[new_dim.transformation_ref]
                xform_events = xform.apply(xform_events)

            new_dims_events.append(xform_events)

        return np.hstack(new_dims_events)

    @staticmethod
    def _apply_parent_results(sample, gate, results, parent_results):
        if parent_results is not None:
            results_and_parent = np.logical_and(parent_results['events'], results)
            parent_count = parent_results['count']
        else:
            # no parent, so results are unchanged & parent count is total count
            parent_count = sample.event_count
            results_and_parent = results

        if isinstance(gate, fk_gates.QuadrantGate):
            final_results = {}

            for q_id, q_result in results_and_parent.items():
                q_event_count = q_result.sum()

                final_results[q_id] = {
                    'sample': sample.original_filename,
                    'events': q_result,
                    'count': q_event_count,
                    'absolute_percent': (q_event_count / float(sample.event_count)) * 100.0,
                    'relative_percent': (q_event_count / float(parent_count)) * 100.0,
                    'parent': gate.parent,
                    'gate_type': gate.gate_type
                }
        else:
            event_count = results_and_parent.sum()

            # check parent_count to avoid div by zero
            if parent_count == 0:
                relative_percent = 0.0
            else:
                relative_percent = (event_count / float(parent_count)) * 100.0
            final_results = {
                'sample': sample.original_filename,
                'events': results_and_parent,
                'count': event_count,
                'absolute_percent': (event_count / float(sample.event_count)) * 100.0,
                'relative_percent': relative_percent,
                'parent': gate.parent,
                'gate_type': gate.gate_type
            }

        return final_results

    def gate_sample(self, sample, cache_events=False, verbose=False):
        """
        Apply GatingStrategy to a Sample instance, returning a GatingResults instance.

        :param sample: an FCS Sample instance
        :param cache_events: Whether to cache pre-processed events (compensated and transformed). This can
            be useful to speed up processing of gates that share the same pre-processing instructions for
            the same channel data, but can consume significantly more memory space. See the related
            clear_cache method for additional information. Default is False.
        :param verbose: If True, print a line for each gate processed
        :return: GatingResults instance
        """
        results = {}

        # The goal here is to avoid re-analyzing any gates.
        # For every gate processed, the results will be cached.
        # Since the nodes are retrieved in hierarchy order,
        # we simply check our cached results for the gate's
        # parent results. This should avoid having any gate
        # processed more than once.
        process_order = list(nx.algorithms.topological_sort(self._dag))

        process_order.remove(('root',))

        for item in process_order:
            # to make the dict key unique, make a string from the ancestors,
            # but also add the set of them as a set to avoid repeated & fragile
            # string splitting in the GatingResults class
            g_id = item[-1]
            g_path = item[:-1]
            gate_path_str = "/".join(g_path)

            p_id = g_path[-1]
            p_path = g_path[:-1]
            parent_gate_path_str = "/".join(p_path)

            g_uid = (g_id, gate_path_str)
            p_uid = (p_id, parent_gate_path_str)

            # check if this gate has already been processed
            if g_uid in results:
                continue

            gate = self.get_gate(g_id, g_path)

            # get_gate returns QuadrantGate when given on of its quadrant, let's check if we have
            # just a quadrant or the full QuadrantGate
            if isinstance(gate, fk_gates.QuadrantGate):
                if g_id in gate.quadrants.keys():
                    # This is a quadrant sub-gate, we'll process the quadrant sub-gates
                    # all at once with the main QuadrantGate ID
                    continue

            if verbose:
                print("%s: processing gate %s" % (sample.original_filename, g_id))

            # look up parent results
            parent_results = None  # default to None
            if gate.parent is not None:
                parent_gate = self.get_gate(p_id, p_path)
                if p_uid in results:
                    parent_results = results[p_uid]
                elif isinstance(parent_gate, fk_gates.QuadrantGate):
                    # need to check for quadrant results in a quadrant gate
                    q_gate_name = p_path[-1]
                    q_gate_path = p_path[:-1]
                    q_gate_res_key = (q_gate_name, "/".join(q_gate_path))
                    if q_gate_res_key in results:
                        parent_results = results[q_gate_res_key][p_id]

            # Call appropriate gate apply method
            if isinstance(gate, fk_gates.BooleanGate):
                # BooleanGate is a bit different, needs a DataFrame of gate ref results
                bool_gate_ref_results = {}
                for gate_ref in gate.gate_refs:
                    gate_ref_gate = self.get_gate(gate_ref['ref'], gate_ref['path'])
                    gate_ref_res_key = (gate_ref['ref'], "/".join(gate_ref['path']))

                    if isinstance(gate_ref_gate, fk_gates.QuadrantGate):
                        quad_gate_name = gate_ref_gate.gate_name
                        quad_gate_path = gate_ref['path'][:-1]

                        quad_gate_res_key = (quad_gate_name, "/".join(quad_gate_path))
                        quad_gate_results = results[quad_gate_res_key]

                        # but the quadrant result is what we're after
                        bool_gate_ref_results[gate_ref_res_key] = quad_gate_results[gate_ref['ref']]['events']
                    else:
                        bool_gate_ref_results[gate_ref_res_key] = results[gate_ref_res_key]['events']

                gate_results = gate.apply(pd.DataFrame(bool_gate_ref_results))
            else:
                # all other gate types process our labelled event DataFrame
                # first, preprocess events then apply gate
                df_events = self._preprocess_sample_events(sample, gate, cache_events=cache_events)
                gate_results = gate.apply(df_events)

            # Combine gate and parent results
            results[g_id, gate_path_str] = self._apply_parent_results(sample, gate, gate_results, parent_results)

            if isinstance(gate, fk_gates.QuadrantGate):
                for quad_res in results[g_id, gate_path_str].values():
                    quad_res['gate_path'] = g_path
            else:
                results[g_id, gate_path_str]['gate_path'] = g_path

        if not cache_events:
            self.clear_cache()

        return GatingResults(results, sample_id=sample.original_filename)



"""
Configuration & system settings used by FlowKit
"""
import sys
import platform

# Used to detect PyCharm's debugging mode to turn off multi-processing for debugging with tests
get_trace = getattr(sys, 'gettrace', lambda: None)

if get_trace() is None:
    debug = False
else:
    debug = True

    if 'coverage' in sys.modules.keys():
        # if running with coverage, don't use debug mode
        debug = False

_platform = platform.system().lower()

if _platform in ['linux', 'darwin']:
    mp_context = 'fork'
else:
    # fork not available on Windows
    mp_context = 'spawn'

# if for any reason multiprocessing is not available, turn it off
try:
    import multiprocessing as mp
    multi_proc = True
except ImportError:
    mp = None
    multi_proc = False



"""
Module containing FlowJo compatible versions of gate classes
"""
import numpy as np
from .. import gates
from ..._utils import gate_utils, xml_utils


class WSPEllipsoidGate(gates.EllipsoidGate):
    """
    Represents a FlowJo workspace Ellipsoid Gate

    FlowJo ellipsoids are odd in that they specify the foci and 4 edge points on the
    ellipse (not the main vertices or co-vertices). They also store coordinates in
    the FlowJo "display space" of 256 x 256. The display space is binned after the
    transformation has been applied, so the multiplier to scale the coordinates is
    found by dividing the max transformed dimension range by 256.
    """
    def __init__(
            self,
            gate_element,
            gating_namespace,
            data_type_namespace
    ):
        gate_name, parent_gate_name, dimensions = xml_utils.parse_gate_element(
            gate_element,
            gating_namespace,
            data_type_namespace
        )

        # First, check the dimensions for transformation references
        # as we need to scale the display space coords to data space.
        # Currently, only supporting linear transformation or None
        for dim in dimensions:
            if dim.transformation_ref is not None:
                raise NotImplementedError("FlowJo ellipses with transformations is not yet supported")

        inv_display_scale = 262144. / 256.

        # Get the foci points of the ellipse, contained in
        # a 'foci' element, that holds 2 'vertex' elements
        foci_el = gate_element.find(
            '%s:foci' % gating_namespace,
            namespaces=gate_element.nsmap
        )

        foci = []

        foci_vertex_els = foci_el.findall(
            '%s:vertex' % gating_namespace,
            namespaces=gate_element.nsmap
        )

        if len(foci_vertex_els) <= 1:
            raise ValueError(
                'Ellipsoids must have at least 2 dimensions (line %d)' % gate_element.sourceline
            )

        for foci_vertex_el in foci_vertex_els:
            coord_els = foci_vertex_el.findall(
                '%s:coordinate' % gating_namespace,
                namespaces=gate_element.nsmap
            )

            coordinates = []

            for coord_el in coord_els:
                value = xml_utils.find_attribute_value(coord_el, data_type_namespace, 'value')
                if value is None:
                    raise ValueError(
                        'A coordinate must have only 1 value (line %d)' % coord_el.sourceline
                    )

                coordinates.append(float(value) * inv_display_scale)

            foci.append(np.array(coordinates))

        # Next, we'll parse the edge vertices
        edge_el = gate_element.find(
            '%s:edge' % gating_namespace,
            namespaces=gate_element.nsmap
        )

        edge_vertices = []

        edge_vertex_els = edge_el.findall(
            '%s:vertex' % gating_namespace,
            namespaces=gate_element.nsmap
        )

        if len(edge_vertex_els) < 4:
            raise ValueError(
                'FlowJo ellipsoids must have 4 edge points (line %d)' % gate_element.sourceline
            )

        for edge_vertex_el in edge_vertex_els:
            coord_els = edge_vertex_el.findall(
                '%s:coordinate' % gating_namespace,
                namespaces=gate_element.nsmap
            )

            coordinates = []

            for coord_el in coord_els:
                value = xml_utils.find_attribute_value(coord_el, data_type_namespace, 'value')
                if value is None:
                    raise ValueError(
                        'A coordinate must have only 1 value (line %d)' % coord_el.sourceline
                    )

                coordinates.append(float(value) * inv_display_scale)

            edge_vertices.append(np.array(coordinates))

        center = (foci[0] + foci[1]) / 2.0

        # need to rotate points on ellipse
        slope = (foci[1][1] - foci[0][1]) / (foci[1][0] - foci[0][0])
        theta_rad = np.arctan(slope)
        cos, sin = np.cos(theta_rad), np.sin(theta_rad)
        r = np.array(((cos, -sin), (sin, cos)))

        rv1 = gate_utils.rotate_point_around_point(edge_vertices[0], r, center)
        rv3 = gate_utils.rotate_point_around_point(edge_vertices[2], r, center)

        # (((x - cx) ** 2) / a ** 2) + (((y - cy) ** 2) / b ** 2) = 1
        # let:
        #     m = 1 / a^2
        #     n = 1 / b^2
        #     xd2 = (x - cx)^2
        #     yd2 = (y - cy)^2
        #
        # then:
        # xd2 * m + yd2 * n = a
        tv1 = rv1 - center
        tv3 = rv3 - center

        a = [tv1 ** 2, tv3 ** 2]
        b = [1., 1.]

        m, n = np.linalg.solve(a, b)
        w = 2 * np.sqrt(1. / abs(m))
        h = 2 * np.sqrt(1. / abs(n))

        # calculate covariance matrix
        w_h_array = np.array([w, h])
        eig_values = (w_h_array / 2.) ** 2

        inv_r = np.linalg.inv(r)
        diag_mat = np.diag(eig_values)
        cov_mat = r.dot(diag_mat).dot(inv_r)
        distance_square = 1.

        super().__init__(
            gate_name,
            parent_gate_name,
            dimensions,
            center,
            cov_mat,
            distance_square
        )



"""
Utility functions related to plotting
"""
import numpy as np
from scipy.interpolate import interpn
import colorsys
from matplotlib import cm, colors
from bokeh.plotting import figure
from bokeh.models import Ellipse, Patch, Span, BoxAnnotation, Rect, ColumnDataSource


line_color = "#1F77B4"
line_color_contrast = "#73D587"
line_width = 3
fill_color = 'lime'
fill_alpha = 0.08


def _generate_custom_colormap(colormap_sample_indices, base_colormap):
    x = np.linspace(0, np.pi, base_colormap.N)
    new_lum = (np.sin(x) * 0.75) + .25

    new_color_list = []

    for i in colormap_sample_indices:
        (r, g, b, a) = base_colormap(i)
        (h, s, v) = colorsys.rgb_to_hsv(r, g, b)

        mod_v = (v * ((196 - abs(i - 196)) / 196) + new_lum[i]) / 2.

        new_r, new_g, new_b = colorsys.hsv_to_rgb(h, 1., mod_v)
        (_, new_l, _) = colorsys.rgb_to_hls(new_r, new_g, new_b)

        new_color_list.append((new_r, new_g, new_b))

    return colors.LinearSegmentedColormap.from_list(
        'custom_' + base_colormap.name,
        new_color_list,
        256
    )


cm_sample = [
    0, 8, 16, 24, 32, 40, 48, 52, 60, 64, 72, 80, 92,
    100, 108, 116, 124, 132,
    139, 147, 155, 159,
    163, 167, 171, 175, 179, 183, 187, 191, 195, 199, 215, 231, 239
]

new_jet = _generate_custom_colormap(cm_sample, cm.get_cmap('jet'))


def _get_false_bounds(bool_array):
    diff = np.diff(np.hstack((0, bool_array, 0)))

    start = np.where(diff == 1)
    end = np.where(diff == -1)

    return start[0], end[0]


def plot_channel(channel_events, label, subplot_ax, xform=None, flagged_events=None):
    """
    Plots a single-channel of FCS event data with the x-axis as the event number (similar to having
    time on the x-axis, but events are equally spaced). This function takes a Matplotlib Axes object
    to enable embedding multiple channel plots within the same figure (created outside this function).

    :param channel_events: 1-D NumPy array of event data
    :param label: string to use as the plot title
    :param subplot_ax: Matplotlib Axes instance used to render the plot
    :param xform: an optional Transform instance used to transform the given event data. channel_events can
        be given already pre-processed (compensated and/or transformed), in this case set xform to None.
    :param flagged_events: optional Boolean array of "flagged" events, regions of flagged events will
        be highlighted in red if flagged_events is given.
    :return: None
    """
    if xform:
        channel_events = xform.apply(channel_events)

    bins = int(np.sqrt(channel_events.shape[0]))
    event_range = range(0, channel_events.shape[0])

    subplot_ax.set_title(label, fontsize=16)
    subplot_ax.set_xlabel("Events", fontsize=14)

    subplot_ax.hist2d(
        event_range,
        channel_events,
        bins=[bins, 128],
        cmap='rainbow',
        cmin=1
    )

    if flagged_events is not None:
        starts, ends = _get_false_bounds(flagged_events)

        for i, s in enumerate(starts):
            subplot_ax.axvspan(
                event_range[s],
                event_range[ends[i] - 1],
                facecolor='pink',
                alpha=0.3,
                edgecolor='deeppink'
            )


def _calculate_extent(data_1d, d_min=None, d_max=None, pad=0.0):
    data_min = data_1d.min()
    data_max = data_1d.max()

    # determine padding to keep min/max events off the edge
    pad_d = max(abs(data_1d.min()), abs(data_1d.max())) * pad

    if d_min is None:
        d_min = data_min - pad_d
    if d_max is None:
        d_max = data_max + pad_d

    return d_min, d_max


def render_polygon(vertices):
    """
    Renders a Bokeh polygon for plotting
    :param vertices: list of 2-D coordinates representing vertices of the polygon
    :return: tuple containing the Bokeh ColumnDataSource and polygon glyphs (as Patch object)
    """
    x_coords, y_coords = list(zip(*[v.coordinates for v in vertices]))

    source = ColumnDataSource(dict(x=x_coords, y=y_coords))

    poly = Patch(
        x='x',
        y='y',
        fill_color=fill_color,
        fill_alpha=fill_alpha,
        line_width=line_width,
        line_color=line_color_contrast
    )

    return source, poly


def render_ranges(dim_minimums, dim_maximums):
    """
    Renders Bokeh Span & BoxAnnotation objects for plotting simple range gates, essentially divider lines.
    There should be no more than 3 items total between dim_minimums & dim_maximums, else the object should
    be rendered as a rectangle.

    :param dim_minimums: list of minimum divider values (max of 2)
    :param dim_maximums: list of maximum divider values (max of 2)
    :return: tuple of Span objects for every item in dim_minimums & dim_maximums
    """
    renderers = []
    left = None
    right = None
    bottom = None
    top = None

    if dim_minimums[0] is not None:
        left = dim_minimums[0]
        renderers.append(
            Span(location=left, dimension='height', line_width=line_width, line_color=line_color)
        )
    if dim_maximums[0] is not None:
        right = dim_maximums[0]
        renderers.append(
            Span(location=right, dimension='height', line_width=line_width, line_color=line_color)
        )
    if len(dim_minimums) > 1:
        if dim_minimums[1] is not None:
            bottom = dim_minimums[1]
            renderers.append(
                Span(location=bottom, dimension='width', line_width=line_width, line_color=line_color)
            )
        if dim_maximums[1] is not None:
            top = dim_maximums[1]
            renderers.append(
                Span(location=top, dimension='width', line_width=line_width, line_color=line_color)
            )

    mid_box = BoxAnnotation(
        left=left,
        right=right,
        bottom=bottom,
        top=top,
        fill_alpha=fill_alpha,
        fill_color=fill_color
    )
    renderers.append(mid_box)

    return renderers


def render_rectangle(dim_minimums, dim_maximums):
    """
    Renders Bokeh Rect object for plotting a rectangle gate.

    :param dim_minimums: list of 2 values representing the lower left corner of a rectangle
    :param dim_maximums: list of 2 values representing the upper right corner of a rectangle
    :return: Bokeh Rect object
    """
    x_center = (dim_minimums[0] + dim_maximums[0]) / 2.0
    y_center = (dim_minimums[1] + dim_maximums[1]) / 2.0
    x_width = dim_maximums[0] - dim_minimums[0]
    y_height = dim_maximums[1] - dim_minimums[1]
    rect = Rect(
        x=x_center,
        y=y_center,
        width=x_width,
        height=y_height,
        fill_color=fill_color,
        fill_alpha=fill_alpha,
        line_width=line_width
    )

    return rect


def render_dividers(x_locs, y_locs):
    """
    Renders lines for divider boundaries (2-D only)
    :param x_locs: list of divider locations in x-axis
    :param y_locs: list of divider locations in y-axis
    :return: list of Bokeh renderer objects
    """
    renderers = []

    for x_loc in x_locs:
        renderers.append(
            Span(location=x_loc, dimension='height', line_width=line_width, line_color=line_color)
        )
    for y_loc in y_locs:
        renderers.append(
            Span(location=y_loc, dimension='width', line_width=line_width, line_color=line_color)
        )

    return renderers


def render_ellipse(center_x, center_y, covariance_matrix, distance_square):
    """
    Renders a Bokeh Ellipse object given the ellipse center point, covariance, and distance square

    :param center_x: x-coordinate of ellipse center
    :param center_y: y-coordinate of ellipse center
    :param covariance_matrix: NumPy array containing the covariance matrix of the ellipse
    :param distance_square: value for distance square of ellipse
    :return: Bokeh Ellipse object
    """
    values, vectors = np.linalg.eigh(covariance_matrix)
    order = values.argsort()[::-1]
    values = values[order]
    vectors = vectors[:, order]

    angle_rads = np.arctan2(*vectors[:, 0][::-1])

    # Width and height are full width (the axes lengths are thus multiplied by 2.0 here)
    width, height = 2.0 * np.sqrt(values * distance_square)

    ellipse = Ellipse(
        x=center_x,
        y=center_y,
        width=width,
        height=height,
        angle=angle_rads,
        line_width=line_width,
        line_color=line_color,
        fill_color=fill_color,
        fill_alpha=fill_alpha
    )

    return ellipse


def plot_histogram(x, x_label='x', bins=None):
    """
    Creates a Bokeh histogram plot of the given 1-D data array.

    :param x: 1-D array of data values
    :param x_label: Label to use for the x-axis
    :param bins: Number of bins to use for the histogram or a string compatible
            with the NumPy histogram function. If None, the number of bins is
            determined by the square root rule.
    :return: Bokeh Figure object containing the histogram
    """
    if bins is None:
        bins = 'sqrt'

    hist, edges = np.histogram(x, density=False, bins=bins)

    tools = "crosshair,hover,pan,zoom_in,zoom_out,box_zoom,undo,redo,reset,save,"

    p = figure(tools=tools)
    p.title.align = 'center'
    p.quad(
        top=hist,
        bottom=0,
        left=edges[:-1],
        right=edges[1:],
        alpha=0.5
    )

    p.y_range.start = 0
    p.xaxis.axis_label = x_label
    p.yaxis.axis_label = 'Event Count'

    return p


def plot_scatter(
        x,
        y,
        dim_ids=None,
        x_min=None,
        x_max=None,
        y_min=None,
        y_max=None,
        color_density=True
):
    """
    Creates a Bokeh scatter plot from the two 1-D data arrays.

    :param x: 1-D array of data values for the x-axis
    :param y: 1-D array of data values for the y-axis
    :param dim_ids: Labels to use for the x-axis & y-axis, respectively
    :param x_min: Lower bound of x-axis. If None, channel's min value will
        be used with some padding to keep events off the edge of the plot.
    :param x_max: Upper bound of x-axis. If None, channel's max value will
        be used with some padding to keep events off the edge of the plot.
    :param y_min: Lower bound of y-axis. If None, channel's min value will
        be used with some padding to keep events off the edge of the plot.
    :param y_max: Upper bound of y-axis. If None, channel's max value will
        be used with some padding to keep events off the edge of the plot.
    :param color_density: Whether to color the events by density, similar
        to a heat map. Default is True.
    :return: A Bokeh Figure object containing the interactive scatter plot.
    """
    if len(x) > 0:
        x_min, x_max = _calculate_extent(x, d_min=x_min, d_max=x_max, pad=0.02)
    if len(y) > 0:
        y_min, y_max = _calculate_extent(y, d_min=y_min, d_max=y_max, pad=0.02)

    if y_max > x_max:
        radius_dimension = 'y'
        radius = 0.003 * y_max
    else:
        radius_dimension = 'x'
        radius = 0.003 * x_max

    if color_density:
        data, x_e, y_e = np.histogram2d(x, y, bins=[38, 38])
        z = interpn(
            (0.5 * (x_e[1:] + x_e[:-1]), 0.5 * (y_e[1:] + y_e[:-1])),
            data,
            np.vstack([x, y]).T,
            method="splinef2d",
            bounds_error=False
        )
        z[np.isnan(z)] = 0

        # sort by density (z) so the more dense points are on top for better
        # color display
        idx = z.argsort()
        x, y, z = x[idx], y[idx], z[idx]
    else:
        z = np.zeros(len(x))

    colors_array = new_jet(colors.Normalize()(z))
    z_colors = [
        "#%02x%02x%02x" % (int(c[0] * 255), int(c[1] * 255), int(c[2] * 255)) for c in colors_array
    ]

    tools = "crosshair,hover,pan,zoom_in,zoom_out,box_zoom,undo,redo,reset,save,"
    p = figure(
        tools=tools,
        x_range=(x_min, x_max),
        y_range=(y_min, y_max)
    )

    p.xaxis.axis_label = dim_ids[0]
    p.yaxis.axis_label = dim_ids[1]

    p.scatter(
        x,
        y,
        radius=radius,
        radius_dimension=radius_dimension,
        fill_color=z_colors,
        fill_alpha=0.4,
        line_color=None
    )

    return p



"""
Utility functions related to FlowJo Workspace files (.wsp)
"""
import copy
import datetime
import re
import numpy as np
from lxml import etree
from .sample_utils import FCS_STANDARD_KEYWORDS
from .xml_utils import find_attribute_value, _get_xml_type
from .._models.dimension import Dimension
# noinspection PyProtectedMember
from .._models.transforms._matrix import Matrix
# noinspection PyProtectedMember
from .._models.transforms import _transforms, _wsp_transforms
# noinspection PyProtectedMember
from .._models.gates._gates import PolygonGate, RectangleGate
# noinspection PyProtectedMember
from .._models.gates._gml_gates import \
    GMLBooleanGate, \
    GMLQuadrantGate, \
    GMLPolygonGate, \
    GMLRectangleGate
# noinspection PyProtectedMember
from .._models.gates._wsp_gates import WSPEllipsoidGate

wsp_gate_constructor_lut = {
    'RectangleGate': GMLRectangleGate,
    'PolygonGate': GMLPolygonGate,
    'EllipsoidGate': WSPEllipsoidGate,
    'QuadrantGate': GMLQuadrantGate,
    'BooleanGate': GMLBooleanGate
}


def _parse_wsp_compensation(sample_el, transform_ns, data_type_ns):
    # Find spilloverMatrix elements, not sure if there should be just a
    # single matrix or multiple. Going with a single one now since there
    # do not appear to be comp references in the WSP gate elements.
    matrix_els = sample_el.findall(
        '%s:spilloverMatrix' % transform_ns,
        namespaces=sample_el.nsmap
    )

    matrix_els_cnt = len(matrix_els)

    if matrix_els_cnt > 1:
        raise ValueError("Multiple spillover matrices per sample are not supported.")
    elif matrix_els_cnt == 0:
        return None

    matrix_el = matrix_els[0]

    # we'll ignore the non-human readable matrix ID and use the name instead
    matrix_name = matrix_el.attrib['name']
    matrix_prefix = matrix_el.attrib['prefix']
    matrix_suffix = matrix_el.attrib['suffix']

    detectors = []
    matrix_array = []

    params_els = matrix_el.find('%s:parameters' % data_type_ns, namespaces=matrix_el.nsmap)
    param_els = params_els.findall('%s:parameter' % data_type_ns, namespaces=matrix_el.nsmap)
    for param_el in param_els:
        param_name = find_attribute_value(param_el, data_type_ns, 'name')
        detectors.append(param_name)

    spill_els = matrix_el.findall(
        '%s:spillover' % transform_ns,
        namespaces=matrix_el.nsmap
    )

    for spill_el in spill_els:
        matrix_row = []

        coefficient_els = spill_el.findall(
            '%s:coefficient' % transform_ns,
            namespaces=spill_el.nsmap
        )

        for co_el in coefficient_els:
            value = find_attribute_value(co_el, transform_ns, 'value')
            if value is None:
                raise ValueError(
                    'Matrix coefficient must have only 1 value (line %d)' % co_el.sourceline
                )

            matrix_row.append(float(value))

        matrix_array.append(matrix_row)

    matrix_array = np.array(matrix_array)
    matrix = Matrix(
        matrix_name,
        matrix_array,
        detectors=detectors,
        fluorochromes=['' for _ in detectors]
    )

    matrix_dict = {
        'matrix_name': matrix_name,
        'prefix': matrix_prefix,
        'suffix': matrix_suffix,
        'detectors': detectors,
        'matrix_array': matrix_array,
        'matrix': matrix
    }

    return matrix_dict


def _parse_wsp_transforms(transforms_el, transform_ns, data_type_ns):
    # get all children and then determine the tag based on the xform type (linear, fasinh, etc.)
    xform_els = transforms_el.getchildren()

    # there should be one transform per channel, use the channel names to create a LUT
    xforms_lut = {}

    for xform_el in xform_els:
        xform_type = xform_el.tag.partition('}')[-1]

        param_el = xform_el.find(
            '%s:parameter' % data_type_ns,
            namespaces=xform_el.nsmap
        )
        param_name = find_attribute_value(param_el, data_type_ns, 'name')

        # FlowKit only supports linear, log, and logicle transformations in FlowJo WSP files.
        # All other bi-ex transforms implemented by FlowJo are undocumented and not reproducible
        if xform_type == 'linear':
            min_range = find_attribute_value(xform_el, transform_ns, 'minRange')
            max_range = find_attribute_value(xform_el, transform_ns, 'maxRange')
            xforms_lut[param_name] = _transforms.LinearTransform(
                param_name,
                param_t=float(max_range),
                param_a=float(min_range)
            )
        elif xform_type == 'log':
            offset = find_attribute_value(xform_el, transform_ns, 'offset')
            decades = find_attribute_value(xform_el, transform_ns, 'decades')
            xforms_lut[param_name] = _wsp_transforms.WSPLogTransform(
                param_name,
                offset=float(offset),
                decades=float(decades)
            )
        elif xform_type == 'logicle':
            # logicle transform has 4 parameters: T, W, M, and A
            # these are attributes of the 'logicle' element
            param_t = find_attribute_value(xform_el, transform_ns, 'T')
            param_w = find_attribute_value(xform_el, transform_ns, 'W')
            param_m = find_attribute_value(xform_el, transform_ns, 'M')
            param_a = find_attribute_value(xform_el, transform_ns, 'A')
            xforms_lut[param_name] = _transforms.LogicleTransform(
                param_name,
                param_t=float(param_t),
                param_w=float(param_w),
                param_m=float(param_m),
                param_a=float(param_a)
            )
        elif xform_type == 'biex':
            # biex transform has 5 parameters, but only 2 are really used
            # these are attributes of the 'biex' element
            param_neg = find_attribute_value(xform_el, transform_ns, 'neg')
            param_width = find_attribute_value(xform_el, transform_ns, 'width')
            param_length = find_attribute_value(xform_el, transform_ns, 'length')
            param_max_range = find_attribute_value(xform_el, transform_ns, 'maxRange')
            param_pos = find_attribute_value(xform_el, transform_ns, 'pos')
            param_pos = round(float(param_pos), 2)

            if param_length != '256':
                raise ValueError("FlowJo biex 'length' parameter value of %s is not supported." % param_length)

            xforms_lut[param_name] = _wsp_transforms.WSPBiexTransform(
                param_name,
                negative=float(param_neg),
                width=float(param_width),
                positive=float(param_pos),
                max_value=float(param_max_range)
            )
        elif xform_type == 'fasinh':
            # FlowJo's implementation of fasinh is slightly different from GML,
            # and uses an additional 'length' scale factor. However, this scaling
            # doesn't seem to affect the results and we can use the regular
            # GML version of asinh. The xform_el also contains other
            # unnecessary parameters: 'length', 'maxRange', and 'W'
            param_t = find_attribute_value(xform_el, transform_ns, 'T')
            param_a = find_attribute_value(xform_el, transform_ns, 'A')
            param_m = find_attribute_value(xform_el, transform_ns, 'M')
            xforms_lut[param_name] = _transforms.AsinhTransform(
                param_name,
                param_t=float(param_t),
                param_m=float(param_m),
                param_a=float(param_a)
            )
        else:
            error_msg = "FlowJo transform type '%s' is undocumented and not supported in FlowKit. " % xform_type
            error_msg += "Please edit the workspace in FlowJo and save all channel transformations as either " \
                "linear, log, biex, logicle, or ArcSinh"

            raise ValueError(error_msg)

    return xforms_lut


def _convert_wsp_gate(wsp_gate, comp_matrix, xform_lut, ignore_transforms=False):
    new_dims = []
    xforms = []

    for dim in wsp_gate.dimensions:
        if comp_matrix is not None:
            pre = comp_matrix['prefix']
            suf = comp_matrix['suffix']
            dim_id = dim.id

            if dim_id.startswith(pre):
                dim_id = re.sub(r'^%s' % pre, '', dim_id)
            if dim_id.endswith(suf):
                dim_id = re.sub(r'%s$' % suf, '', dim_id)

            if dim_id in comp_matrix['detectors']:
                comp_ref = comp_matrix['matrix_name']
            else:
                comp_ref = None
        else:
            dim_id = dim.id
            comp_ref = None

        xform_id = None
        new_dim_min = None
        new_dim_max = None

        if dim_id in xform_lut and not ignore_transforms:
            xform = xform_lut[dim_id]
            xforms.append(xform)  # need these later for vertices, coordinates, etc.
            xform_id = xform.id
            if dim.min is not None:
                new_dim_min = xform.apply(np.array([[float(dim.min)]]))

            if dim.max is not None:
                new_dim_max = xform.apply(np.array([[float(dim.max)]]))
        else:
            xforms.append(None)
            if dim.min is not None:
                new_dim_min = float(dim.min)

            if dim.max is not None:
                new_dim_max = float(dim.max)

        new_dim = Dimension(
            dim_id,
            comp_ref,
            xform_id,
            range_min=new_dim_min,
            range_max=new_dim_max
        )
        new_dims.append(new_dim)

    if isinstance(wsp_gate, GMLPolygonGate):
        # convert vertices using saved xforms
        vertices = copy.deepcopy(wsp_gate.vertices)
        for v in vertices:
            for i, c in enumerate(v.coordinates):
                if xforms[i] is not None:
                    v.coordinates[i] = xforms[i].apply(np.array([[float(c)]]))[0][0]

        gate = PolygonGate(wsp_gate.gate_name, wsp_gate.parent, new_dims, vertices)
    elif isinstance(wsp_gate, GMLRectangleGate):
        gate = copy.deepcopy(wsp_gate)
        gate.dimensions = new_dims
    elif isinstance(wsp_gate, WSPEllipsoidGate):
        gate = copy.deepcopy(wsp_gate)
    else:
        raise NotImplemented(
            "%s gates for FlowJo workspaces are not currently supported." % type(wsp_gate).__name__
        )

    return gate


def _recurse_wsp_sub_populations(sub_pop_el, gate_path, gating_ns, data_type_ns):
    gates = []
    ns_map = sub_pop_el.nsmap

    if gate_path is None:
        gate_path = []
        parent_id = None
    else:
        parent_id = gate_path[-1]

    pop_els = sub_pop_el.findall('Population', ns_map)
    for pop_el in pop_els:
        pop_name = pop_el.attrib['name']
        owning_group = pop_el.attrib['owningGroup']
        gate_el = pop_el.find('Gate', ns_map)

        gate_child_els = gate_el.getchildren()

        if len(gate_child_els) != 1:
            raise ValueError("Gate element must have only 1 child element")

        gate_child_el = gate_child_els[0]

        # determine gate type
        # TODO: this string parsing seems fragile, may need to be shored up
        gate_type = gate_child_el.tag.partition('}')[-1]
        gate_class = wsp_gate_constructor_lut[gate_type]

        g = gate_class(
            gate_child_el,
            gating_ns,
            data_type_ns
        )

        # replace ID and parent ID with population names, but save the originals
        # so we can re-create the correct hierarchy later.
        # NOTE: Some wsp files have the ID as an attribute of the GatingML-style sub-element,
        #       though it isn't always the case. However, it seems the ID is reliably included
        #       in the parent "Gate" element, saved here in the 'gate_el' variable.
        #       Likewise for parent_id
        g.gate_name = pop_name
        g.parent = parent_id

        gates.append(
            {
                'owning_group': owning_group,
                'gate': g,
                'gate_path': gate_path
            }
        )

        sub_pop_els = pop_el.findall('Subpopulations', ns_map)
        child_gate_path = copy.copy(gate_path)
        child_gate_path.append(pop_name)
        for el in sub_pop_els:
            gates.extend(_recurse_wsp_sub_populations(el, child_gate_path, gating_ns, data_type_ns))

    return gates


def _parse_wsp_groups(group_node_els, ns_map, gating_ns, data_type_ns):
    wsp_groups = {}

    for group_node_el in group_node_els:
        group_name = group_node_el.attrib['name']
        group_samples = []
        group_gates = []

        # Group membership for samples is found in the 'Group' branch.
        # The SampleRefs element will contain a SampleRef element for
        # each sample assigned to the group.
        group_el = group_node_el.find('Group', ns_map)
        if group_el is not None:
            group_sample_refs_el = group_el.find('SampleRefs', ns_map)

            if group_sample_refs_el is not None:
                group_sample_ref_els = group_sample_refs_el.findall('SampleRef', ns_map)
                for s_ref_el in group_sample_ref_els:
                    group_s_id = s_ref_el.attrib['sampleID']
                    group_samples.append(group_s_id)

        group_root_sub_pop_el = group_node_el.find('Subpopulations', ns_map)

        # ignore groups with no sub-populations
        if group_root_sub_pop_el is not None:
            group_gates = _recurse_wsp_sub_populations(
                group_root_sub_pop_el,
                None,  # starting at root, so no gate path
                gating_ns,
                data_type_ns
            )

        wsp_groups[group_name] = {
            'gates': group_gates,
            'samples': group_samples
        }

    return wsp_groups


def _parse_wsp_samples(sample_els, ns_map, gating_ns, transform_ns, data_type_ns):
    wsp_samples = {}

    for sample_el in sample_els:
        transforms_el = sample_el.find('Transformations', ns_map)
        sample_node_el = sample_el.find('SampleNode', ns_map)
        sample_name = sample_node_el.attrib['name']
        sample_id = sample_node_el.attrib['sampleID']

        # It appears there is only a single set of xforms per sample, one for each channel.
        # And, the xforms have no IDs. We'll extract it and give it IDs based on ???
        sample_xform_lut = _parse_wsp_transforms(transforms_el, transform_ns, data_type_ns)

        # parse spilloverMatrix elements
        sample_comp = _parse_wsp_compensation(sample_el, transform_ns, data_type_ns)

        # FlowJo WSP gates are nested so we'll have to do a recursive search from the root
        # Sub-populations node
        sample_root_sub_pop_el = sample_node_el.find('Subpopulations', ns_map)

        if sample_root_sub_pop_el is None:
            continue

        sample_gates = _recurse_wsp_sub_populations(
            sample_root_sub_pop_el,
            None,  # starting at root, so no gate path
            gating_ns,
            data_type_ns
        )

        # sample gate LUT will store everything we need to convert sample gates,
        # including any custom gates (ones with empty string owning groups).
        wsp_samples[sample_id] = {
            'sample_name': sample_name,
            'custom_gate_ids': set(),
            'custom_gates': [],
            'transforms': sample_xform_lut,
            'comp': sample_comp
        }

        for sample_gate in sample_gates:
            if sample_gate['owning_group'] == '':
                # If the owning group is an empty string, it is a custom gate for that sample
                # that is potentially used in another group. However, it appears that if a
                # sample has a custom gate then that custom gate cannot be further customized.
                # Since there is only a single custom gate per gate name per sample, then we
                # can create a LUT of custom gates per sample
                wsp_samples[sample_id]['custom_gate_ids'].add(sample_gate['gate'].gate_name)
                wsp_samples[sample_id]['custom_gates'].append(
                    {
                        'gate': sample_gate['gate'],
                        'gate_path': sample_gate['gate_path']
                    }
                )

    return wsp_samples


def parse_wsp(workspace_file_or_path, ignore_transforms=False):
    """
    Converts a FlowJo 10 workspace file (.wsp) into a nested Python dictionary with the following structure:

        wsp_dict[group][sample_name] = {
            'gates': {gate_id: gate_instance, ...},
            'transforms': transforms_list,
            'compensation': matrix
        }

    :param workspace_file_or_path: A FlowJo .wsp file or file path
    :param ignore_transforms: Some FlowJo 10 transforms are incompatible with FlowKit. This boolean argument
        allows parsing workspace files with unsupported transforms without raising errors. Default is False.
    :return: dict
    """
    doc_type, root_xml, gating_ns, data_type_ns, transform_ns = _get_xml_type(workspace_file_or_path)

    # first, find 1st level elements:
    #     - Groups -> GroupNode
    #     - SampleList -> Sample
    ns_map = root_xml.nsmap
    groups_el = root_xml.find('Groups', ns_map)
    group_node_els = groups_el.findall('GroupNode', ns_map)
    sample_list_el = root_xml.find('SampleList', ns_map)
    sample_els = sample_list_el.findall('Sample', ns_map)

    # Find all group gates before looking a the custom sample gates.
    # Custom sample gates are defined in the SampleList branch and
    # are indicated by having an empty string for the owningGroup.
    # Only 1 custom sample gate can exist per sample, regardless
    # of sample group. So, if a group contains a gate with a name
    # that matches a custom sample gate name, then the custom one
    # overrides the group gate.
    wsp_groups = _parse_wsp_groups(group_node_els, ns_map, gating_ns, data_type_ns)

    # Now parse the Sample elements in SampleList looking for custom sample gates
    wsp_samples = _parse_wsp_samples(sample_els, ns_map, gating_ns, transform_ns, data_type_ns)

    # Now we can assemble the complete gating strategies but the gates
    # need to be converted to the transformed space according to the
    # correct specified transforms. This is because FlowJo gates are
    # stored in the untransformed space.
    wsp_dict = {}

    for group_id, group_dict in wsp_groups.items():
        wsp_dict[group_id] = {}

        for group_sample_id in group_dict['samples']:
            sample_dict = wsp_samples[group_sample_id]
            sample_name = sample_dict['sample_name']

            group_sample_gate_names = []
            group_sample_gates = []

            for group_gate in group_dict['gates']:
                group_gate_name = group_gate['gate'].gate_name

                tmp_gate = copy.deepcopy(group_gate['gate'])

                if group_gate_name in sample_dict['custom_gate_ids']:
                    group_gate_path = group_gate['gate_path']
                    for sample_gate_dict in sample_dict['custom_gates']:
                        # noinspection PyTypeChecker
                        tmp_sample_gate = sample_gate_dict['gate']
                        # noinspection PyTypeChecker
                        tmp_sample_gate_path = sample_gate_dict['gate_path']
                        # noinspection PyUnresolvedReferences
                        if group_gate_path == tmp_sample_gate_path and tmp_sample_gate.gate_name == group_gate_name:
                            # found a match, overwrite tmp_gate
                            tmp_gate = tmp_sample_gate

                tmp_gate = _convert_wsp_gate(
                    tmp_gate,
                    sample_dict['comp'],
                    sample_dict['transforms'],
                    ignore_transforms=ignore_transforms
                )

                group_sample_gate_names.append(group_gate_name)
                group_sample_gates.append(
                    {
                        'gate': tmp_gate,
                        'gate_path': group_gate['gate_path']
                    }
                )

            # Now, we need to check if there were only custom sample gates
            # and no group gates. In this case the above would never have
            # found the custom sample gates, but we don't want to replicate
            # them.
            for sample_gate_dict in sample_dict['custom_gates']:
                # noinspection PyTypeChecker
                sample_gate = sample_gate_dict['gate']
                # noinspection PyTypeChecker
                sample_gate_path = sample_gate_dict['gate_path']

                # noinspection PyUnresolvedReferences
                if sample_gate.gate_name not in group_sample_gate_names:
                    # noinspection PyTypeChecker
                    tmp_gate = _convert_wsp_gate(
                        sample_gate,
                        sample_dict['comp'],
                        sample_dict['transforms'],
                        ignore_transforms=ignore_transforms
                    )
                    group_sample_gates.append(
                        {
                            'gate': tmp_gate,
                            'gate_path': sample_gate_path
                        }
                    )

            if sample_dict['comp'] is not None:
                matrix = sample_dict['comp']['matrix']
            else:
                matrix = None

            wsp_dict[group_id][sample_name] = {
                'gates': group_sample_gates,
                'transforms': list(sample_dict['transforms'].values()),
                'compensation': matrix
            }

    return wsp_dict


def _add_matrix_to_wsp(parent_el, prefix, matrix, ns_map):
    matrix_el = etree.SubElement(parent_el, "{%s}spilloverMatrix" % ns_map['transforms'])
    matrix_el.set('spectral', "0")
    matrix_el.set('prefix', prefix)
    matrix_el.set('name', matrix.id)
    matrix_el.set('version', "FlowJo-10.7.1")
    # TODO: need to set id?
    matrix_el.set('suffix', "")

    params_el = etree.SubElement(matrix_el, "{%s}parameters" % ns_map['data-type'])

    # add parameter labels & their prefix
    for param_label in matrix.detectors:
        param_el = etree.SubElement(params_el, "{%s}parameter" % ns_map['data-type'])
        param_el.set('{%s}name' % ns_map['data-type'], param_label)
        param_el.set('userProvidedCompInfix', "".join([prefix, param_label]))

    # add the matrix array values
    matrix_array = matrix.matrix
    for i_row, param_label in enumerate(matrix.detectors):
        spillover_el = etree.SubElement(matrix_el, "{%s}spillover" % ns_map['transforms'])
        spillover_el.set('{%s}parameter' % ns_map['data-type'], param_label)
        spillover_el.set('userProvidedCompInfix', "".join([prefix, param_label]))

        for i_col, param_label2 in enumerate(matrix.detectors):
            matrix_value = matrix_array[i_row, i_col]

            coef_el = etree.SubElement(spillover_el, "{%s}coefficient" % ns_map['transforms'])
            coef_el.set('{%s}parameter' % ns_map['data-type'], param_label2)
            coef_el.set('{%s}value' % ns_map['transforms'], str(matrix_value))


def _add_transform_to_wsp(parent_el, parameter_label, transform, ns_map):
    if isinstance(transform, _transforms.LinearTransform):
        xform_el = etree.SubElement(parent_el, "{%s}linear" % ns_map['transforms'])
        xform_el.set('{%s}minRange' % ns_map['transforms'], str(transform.param_a))
        xform_el.set('{%s}maxRange' % ns_map['transforms'], str(transform.param_t))
    elif isinstance(transform, _transforms.LogicleTransform):
        xform_el = etree.SubElement(parent_el, "{%s}logicle" % ns_map['transforms'])
        xform_el.set('{%s}length' % ns_map['transforms'], "256")
        xform_el.set('{%s}T' % ns_map['transforms'], str(transform.param_t))
        xform_el.set('{%s}A' % ns_map['transforms'], str(transform.param_a))
        xform_el.set('{%s}W' % ns_map['transforms'], str(transform.param_w))
        xform_el.set('{%s}M' % ns_map['transforms'], str(transform.param_m))
    else:
        raise NotImplementedError("Transform type %s is not yet supported" % type(transform))

    param_el = etree.SubElement(xform_el, "{%s}parameter" % ns_map['data-type'])
    param_el.set('{%s}name' % ns_map['data-type'], str(parameter_label))


def _add_sample_keywords_to_wsp(parent_el, sample):
    # setup some regex vars for detecting dynamic FCS standard keywords
    # i.e. PnN, PnS, etc.
    regex_pnx = re.compile(r"^p(\d+)[bdefglnoprst]$", re.IGNORECASE)

    # start by adding special FlowJo keyword for the FCS version
    kw_el = etree.SubElement(parent_el, "Keyword")
    kw_el.set('name', "FJ_FCS_VERSION")
    kw_el.set('value', str(sample.version))

    for kw, val in sample.metadata.items():
        fcs_kw = False
        if kw in FCS_STANDARD_KEYWORDS:
            fcs_kw = True
        else:
            match = regex_pnx.match(kw)
            if match:
                fcs_kw = True

        kw_el = etree.SubElement(parent_el, "Keyword")
        if fcs_kw:
            kw_el.set('name', "$%s" % kw.upper())
        else:
            kw_el.set('name', kw.upper())

        kw_el.set('value', val)


def _add_polygon_gate(parent_el, gate, fj_gate_id, fj_parent_gate_id, gating_strategy, comp_prefix, ns_map):
    gate_instance_el = etree.SubElement(parent_el, "{%s}PolygonGate" % ns_map['gating'])
    gate_instance_el.set('quadID', '-1')
    gate_instance_el.set('gateResolution', '256')
    gate_instance_el.set('{%s}id' % ns_map['gating'], "ID%s" % fj_gate_id)
    if fj_parent_gate_id is not None:
        gate_instance_el.set('{%s}parent_id' % ns_map['gating'], "ID%s" % fj_parent_gate_id)

    xform_refs = []
    for dim in gate.dimensions:
        # use comp prefix for label except for scatter and time channels
        if dim.id[:4] in ['FSC-', 'SSC-', 'Time']:
            dim_id = dim.id
        else:
            dim_id = comp_prefix + dim.id

        dim_el = etree.SubElement(gate_instance_el, "{%s}dimension" % ns_map['gating'])
        fcs_dim_el = etree.SubElement(dim_el, "{%s}fcs-dimension" % ns_map['data-type'])
        fcs_dim_el.set("{%s}name" % ns_map['data-type'], dim_id)

        xform_refs.append(dim.transformation_ref)

    for vertex in gate.vertices:
        vertex_el = etree.SubElement(gate_instance_el, "{%s}vertex" % ns_map['gating'])
        for dim_idx, coord in enumerate(vertex.coordinates):
            if xform_refs[dim_idx] is not None:
                xform = gating_strategy.get_transform(xform_refs[dim_idx])
                inv_coord = xform.inverse(np.array([[coord]]))[0, 0]
            else:
                inv_coord = coord

            coord_el = etree.SubElement(vertex_el, "{%s}coordinate" % ns_map['gating'])
            coord_el.set("{%s}value" % ns_map['data-type'], str(inv_coord))


def _add_rectangle_gate(parent_el, gate, fj_gate_id, fj_parent_gate_id, gating_strategy, comp_prefix, ns_map):
    gate_instance_el = etree.SubElement(parent_el, "{%s}RectangleGate" % ns_map['gating'])
    gate_instance_el.set('percentX', '0')
    gate_instance_el.set('percentY', '0')
    gate_instance_el.set('{%s}id' % ns_map['gating'], "ID%s" % fj_gate_id)
    if fj_parent_gate_id is not None:
        gate_instance_el.set('{%s}parent_id' % ns_map['gating'], "ID%s" % fj_parent_gate_id)

    for dim in gate.dimensions:
        # use comp prefix for label except for scatter and time channels
        if dim.id[:4] in ['FSC-', 'SSC-', 'Time']:
            dim_id = dim.id
        else:
            dim_id = comp_prefix + dim.id

        dim_el = etree.SubElement(gate_instance_el, "{%s}dimension" % ns_map['gating'])
        fcs_dim_el = etree.SubElement(dim_el, "{%s}fcs-dimension" % ns_map['data-type'])
        fcs_dim_el.set("{%s}name" % ns_map['data-type'], dim_id)

        xform_ref = dim.transformation_ref

        if xform_ref is not None:
            xform = gating_strategy.get_transform(xform_ref)
            dim_min = xform.inverse(np.array([[dim.min]]))[0, 0]
            dim_max = xform.inverse(np.array([[dim.max]]))[0, 0]
        else:
            dim_min = dim.min
            dim_max = dim.max

        dim_el.set('{%s}min' % ns_map['gating'], str(dim_min))
        dim_el.set('{%s}max' % ns_map['gating'], str(dim_max))


def _add_group_node_to_wsp(parent_el, group_name, sample_id_list):
    group_node_el = etree.SubElement(parent_el, "GroupNode")
    group_node_el.set('name', group_name)
    group_node_el.set('annotation', "")
    group_node_el.set('owningGroup', group_name)

    group_el = etree.SubElement(group_node_el, "Group")
    group_el.set('name', group_name)

    sample_refs_el = etree.SubElement(group_el, "SampleRefs")

    for sample_id in sample_id_list:
        sample_ref_el = etree.SubElement(sample_refs_el, "SampleRef")
        sample_ref_el.set('sampleID', sample_id)


def _recurse_add_sub_populations(parent_el, gate_id, gate_path, gating_strategy, gate_fj_id_lut, comp_prefix, ns_map):
    # first, add given gate to parent XML element inside it's own Population element
    pop_el = etree.SubElement(parent_el, "Population")
    pop_el.set('name', gate_id)
    pop_el.set('annotation', "")
    pop_el.set('owningGroup', "")

    # Lookup the gate's FJ ID
    fj_id = gate_fj_id_lut[(gate_id, tuple(gate_path))]

    # Determine if gate has a parent gate & lookup it's FJ ID
    if len(gate_path) > 1:
        # gate has a true parent gate
        parent_gate_id = gate_path[-1]
        parent_gate_path = gate_path[:-1]
        parent_fj_id = gate_fj_id_lut[(parent_gate_id, tuple(parent_gate_path))]
    else:
        parent_fj_id = None

    gate_el = etree.SubElement(pop_el, 'Gate')
    gate_el.set('{%s}id' % ns_map['gating'], "ID%s" % fj_id)

    if parent_fj_id is not None:
        gate_el.set('{%s}parent_id' % ns_map['gating'], "ID%s" % parent_fj_id)

    # Get the gate instance to determine the gate class
    gate = gating_strategy.get_gate(gate_id, gate_path)

    if isinstance(gate, PolygonGate):
        _add_polygon_gate(gate_el, gate, fj_id, parent_fj_id, gating_strategy, comp_prefix, ns_map)
    elif isinstance(gate, RectangleGate):
        _add_rectangle_gate(gate_el, gate, fj_id, parent_fj_id, gating_strategy, comp_prefix, ns_map)
    else:
        raise NotImplementedError("Exporting %s gates is not yet implemented" % str(gate.__class__))

    # If there are child gates, create a new Sub-pop element and recurse
    child_gates = gating_strategy.get_child_gates(gate_id, gate_path)
    if len(child_gates) > 0:
        sub_pops_el = etree.SubElement(pop_el, "Subpopulations")

        # child gate path will be the parent's gate path plus the parent ID
        child_gate_path = copy.deepcopy(gate_path)
        child_gate_path = child_gate_path + (gate_id,)
        for child_gate in child_gates:
            _recurse_add_sub_populations(
                sub_pops_el,
                child_gate.gate_name,
                child_gate_path,
                gating_strategy,
                gate_fj_id_lut,
                comp_prefix,
                ns_map
            )


def _add_sample_node_to_wsp(parent_el, sample_name, sample_id, gating_strategy, comp_prefix_lut, ns_map):
    sample_node_el = etree.SubElement(parent_el, "SampleNode")
    sample_node_el.set('name', sample_name)
    sample_node_el.set('annotation', "")
    sample_node_el.set('owningGroup', "")
    sample_node_el.set('sampleID', sample_id)

    sub_pops_el = etree.SubElement(sample_node_el, "Subpopulations")

    gate_fj_id = 1
    gate_fj_id_lut = {}
    for gate_id, gate_path in gating_strategy.get_gate_ids():
        gate_fj_id_lut[(gate_id, tuple(gate_path))] = str(sample_id) + str(gate_fj_id)
        gate_fj_id += 1

    root_gates = gating_strategy.get_root_gates()

    # need to find a matching compensation to add the correct comp prefix to parameter labels
    comp_ids = gating_strategy.comp_matrices.keys()

    # there really shouldn't be more than 1 compensation matrix and we only support 1 for now
    comp_prefix = ''
    for comp_id in comp_ids:
        if comp_id in comp_prefix_lut:
            comp_prefix = comp_prefix_lut[comp_id]
            break

    for gate in root_gates:
        _recurse_add_sub_populations(
            sub_pops_el,
            gate.gate_name,
            ('root',),
            gating_strategy,
            gate_fj_id_lut,
            comp_prefix,
            ns_map
        )


def export_flowjo_wsp(group_gating_strategies, group_name, samples, file_handle):
    """
    Exports a FlowJo 10 workspace file (.wsp) from the given GatingStrategy instance
    :param group_gating_strategies: dictionary of Session sample group GatingStrategy instances
    :param group_name: text string label for sample group
    :param samples: list of Sample instances associated with the sample group
    :param file_handle: File handle for exported FlowJo workspace file
    :return: None
    """
    # < Workspace
    # version = "20.0"
    # modDate = "Thu Aug 20 12:19:04 EDT 2020"
    # flowJoVersion = "10.6.2"
    # curGroup = "All Samples"
    # xmlns: xsi = "http://www.w3.org/2001/XMLSchema-instance"
    # xmlns: gating = "http://www.isac-net.org/std/Gating-ML/v2.0/gating"
    # xmlns: transforms = "http://www.isac-net.org/std/Gating-ML/v2.0/transformations"
    # xmlns: data - type = "http://www.isac-net.org/std/Gating-ML/v2.0/datatypes"
    ns_g = "http://www.isac-net.org/std/Gating-ML/v2.0/gating"
    ns_dt = "http://www.isac-net.org/std/Gating-ML/v2.0/datatypes"
    ns_xform = "http://www.isac-net.org/std/Gating-ML/v2.0/transformations"
    ns_map = {
        'gating': ns_g,
        'data-type': ns_dt,
        'transforms': ns_xform
    }

    now = datetime.datetime.now().astimezone()
    mod_date = now.strftime('%a %b %d %H:%M:%S %Z %Y')

    root = etree.Element('Workspace', nsmap=ns_map)
    root.set('version', "20.0")
    root.set('modDate', mod_date)
    root.set('flowJoVersion', "10.6.2")
    root.set('curGroup', "All Samples")

    matrices_el = etree.SubElement(root, "Matrices")
    groups_el = etree.SubElement(root, "Groups")
    sample_list_el = etree.SubElement(root, "SampleList")

    template_strategy = group_gating_strategies['template']
    sample_strategies = group_gating_strategies['samples']

    # For now, we'll assume all the comps are in the template
    comp_prefix_counter = 0
    comp_prefix_lut = {}
    for matrix_id, matrix in template_strategy.comp_matrices.items():
        if comp_prefix_counter == 0:
            comp_prefix = 'Comp-'
        else:
            comp_prefix = 'Comp%d-' % comp_prefix_counter

        comp_prefix_lut[matrix_id] = comp_prefix

        _add_matrix_to_wsp(matrices_el, comp_prefix, matrix, ns_map)

        comp_prefix_counter += 1

    # create FJ sample ID LUT (used in various places)
    curr_sample_id = 1  # each sample needs an ID, we'll start at 1 & increment
    sample_id_lut = {}
    for sample in samples:
        # Store sample ID and increment
        sample_id_lut[sample.original_filename] = str(curr_sample_id)
        curr_sample_id += 1

    _add_group_node_to_wsp(groups_el, group_name, sample_id_lut.values())

    gate_ids = template_strategy.get_gate_ids()
    gates = []
    dim_xform_lut = {}  # keys are dim label, value is a set of xform refs

    # Also assume the xforms for all samples are the sam
    for g_id, g_path in gate_ids:
        gate = template_strategy.get_gate(g_id, g_path)
        gates.append(gate)

        for dim in gate.dimensions:
            if dim.id not in dim_xform_lut.keys():
                dim_xform_lut[dim.id] = set()

            dim_xform_lut[dim.id].add(dim.transformation_ref)

    for dim in dim_xform_lut:
        xform_refs = list(dim_xform_lut[dim])
        if len(xform_refs) > 1:
            raise ValueError(
                "The given GatingStrategy is incompatible with FlowJo. "
                "Multiple transformations are not allowed for the same parameter."
                "Parameter %s has multiple transformation references." % dim
            )

        # After verifying there's only 1 transformation reference, set
        # the value in the LUT to the single reference string
        dim_xform_lut[dim] = xform_refs[0]

    # Build SampleList branch
    # The SampleList element contains a series of Sample elements.
    # Each Sample element contains one of the following elements:
    #     - DataSet
    #     - Transformations
    #     - Keywords
    #     - SampleNode
    for sample in samples:
        sample_id = sample_id_lut[sample.original_filename]
        sample_strategy = sample_strategies[sample.original_filename]

        sample_el = etree.SubElement(sample_list_el, "Sample")

        # Start with DataSet sub-element. We don't have the exact
        # file path to the FCS file, so we'll make a relative
        # path using the Sample's original filename, hoping that
        # FlowJo can re-connect the files using that name.
        data_set_el = etree.SubElement(sample_el, "DataSet")
        data_set_el.set('uri', sample.original_filename)
        data_set_el.set('sampleID', sample_id)

        # Transforms in FlowJo are organized differently than in GatingML
        # or a FlowKit GatingStrategy. Instead of transforms being created
        # and referenced by one or more parameters, FlowJo transforms contain
        # the parameter as a child element. This means they cannot be created
        # and re-used for multiple parameters and that the parameter label
        # must be known prior to creating the transform. So, we must dig into
        # the Dimensions of each gate in the entire hierarchy to find all the
        # parameter labels and their reference FlowKit transform.
        xforms_el = etree.SubElement(sample_el, "Transformations")
        for dim, xform in dim_xform_lut.items():
            if xform is None:
                continue
            _add_transform_to_wsp(xforms_el, dim, sample_strategy.get_transform(xform), ns_map)

            # We also need to add comp-prefixed param transforms (excluding scatter and Time channels)
            if dim[:4] in ['FSC-', 'SSC-', 'Time']:
                continue

            for comp_prefix in comp_prefix_lut.values():
                _add_transform_to_wsp(xforms_el, comp_prefix + dim, sample_strategy.get_transform(xform), ns_map)

        # Add Keywords sub-element using the Sample metadata
        keywords_el = etree.SubElement(sample_el, "Keywords")
        _add_sample_keywords_to_wsp(keywords_el, sample)

        # Finally, add the SampleNode where all the gates are defined
        _add_sample_node_to_wsp(
            sample_el,
            sample.original_filename,
            sample_id,
            sample_strategy,
            comp_prefix_lut,
            ns_map
        )

    et = etree.ElementTree(root)

    et.write(file_handle, encoding="utf-8", xml_declaration=True, pretty_print=True)



"""
Setup script for the FlowKit package
"""
from setuptools import dist, setup, find_packages, Extension

# read in version string
VERSION_FILE = 'flowkit/_version.py'
__version__ = None  # to avoid inspection warning and check if __version__ was loaded
exec(open(VERSION_FILE).read())

if __version__ is None:
    raise RuntimeError("__version__ string not found in file %s" % VERSION_FILE)

# NumPy is needed to build
# This retrieves a version at build time compatible with run time version
dist.Distribution().fetch_build_eggs(['numpy>=1.19'])

# override inspection for import not at top of file
# this has to be imported here, after fetching the NumPy egg
import numpy as np  # noqa: E402

with open("README.md", "r") as fh:
    long_description = fh.read()

utils_extension = Extension(
    'flowkit._utils_c',
    sources=[
        'flowkit/utils_c_ext/_utils.c',
        'flowkit/utils_c_ext/utils.c'
    ],
    include_dirs=[np.get_include(), 'flowkit/utils_c_ext'],
    extra_compile_args=['-std=c99']
)

reqs = [
    'flowio',
    'flowutils',
    'matplotlib',
    'scipy',
    'statsmodels',
    'seaborn',
    'pandas',
    'numpy',
    'lxml',
    'bokeh',
    'anytree',
    'networkx',
    'psutil'
]

setup(
    name='FlowKit',
    version=__version__,
    packages=find_packages(exclude=["flowkit/tests/"]),
    package_data={'': ['_resources/*.xsd']},
    include_package_data=True,
    description='Flow Cytometry Toolkit',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Scott White",
    license='BSD',
    url="https://github.com/whitews/flowkit",
    ext_modules=[utils_extension],
    install_requires=reqs,
    classifiers=[
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.6'
    ]
)



"""
Matrix class
"""
import numpy as np
import pandas as pd
import flowutils


class Matrix(object):
    """
    Represents a single compensation matrix from a CSV/TSV file, NumPy array or pandas
    DataFrame.

    :param matrix_id: Text string used to identify the matrix (cannot be 'uncompensated' or 'fcs')
    :param spill_data_or_file: matrix data array, can be either:
            - text string from FCS $SPILL or $SPILLOVER metadata value
            - a file path or file handle to a CSV/TSF file
            - a pathlib Path object to a CSV/TSF file
            - a NumPy array of spill data
            - a pandas DataFrame (channel labels as headers)
    :param detectors: A list of strings or a list of tuples to use for the detector labels.
    :param fluorochromes: A list of strings or a list of tuples to use for the detector labels.
    :param null_channels: List of PnN labels for channels that were collected
        but do not contain useful data. Note, this should only be used if there were
        truly no fluorochromes used targeting those detectors and the channels
        do not contribute to compensation.
    """
    def __init__(
            self,
            matrix_id,
            spill_data_or_file,
            detectors,
            fluorochromes=None,
            null_channels=None
    ):
        if matrix_id == 'uncompensated' or matrix_id == 'fcs':
            raise ValueError(
                "Matrix IDs 'uncompensated' and 'fcs' are reserved compensation references " +
                "used in Dimension instances to specify that channel data should either be " +
                "uncompensated or compensated using the spill value from a Sample's metadata"
            )

        if isinstance(spill_data_or_file, np.ndarray):
            spill = spill_data_or_file
        else:
            spill = flowutils.compensate.parse_compensation_matrix(
                spill_data_or_file,
                detectors,
                null_channels=null_channels
            )
            spill = spill[1:, :]

        self.id = matrix_id
        self.matrix = spill
        # TODO: Should we use a different name other than 'fluorochromes'? They are typically antibodies or markers.
        self.detectors = detectors
        # Note: fluorochromes attribute is required for compatibility with GatingML exports,
        #       as the GatingML 2.0 requires both the set of detectors and fluorochromes.
        if fluorochromes is None:
            fluorochromes = ['' for _ in detectors]

        self.fluorochomes = fluorochromes

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'{self.id}, dims: {len(self.detectors)})'
        )

    def apply(self, sample):
        """
        Apply compensation matrix to given Sample instance.
        :param sample: Sample instance with matching set of detectors
        :return: NumPy array of compensated events
        """
        indices = [
            sample.get_channel_index(d) for d in self.detectors
        ]
        events = sample.get_events(source='raw')

        return flowutils.compensate.compensate(
            events,
            self.matrix,
            indices
        )

    def inverse(self, sample):
        """
        Apply compensation matrix to given Sample instance.
        :param sample: Sample instance with matching set of detectors
        :return: NumPy array of compensated events
        """
        indices = [
            sample.get_channel_index(d) for d in self.detectors
        ]
        events = sample.get_events(source='comp')

        return flowutils.compensate.inverse_compensate(
            events,
            self.matrix,
            indices
        )

    def as_dataframe(self, fluoro_labels=False):
        """
        Returns the compensation matrix as a pandas DataFrame.

        :param fluoro_labels: If True, the fluorochrome names are used as the column headers & row indices, else
            the detector names are used (default).
        :return: pandas DataFrame
        """
        if fluoro_labels:
            labels = self.fluorochomes
        else:
            labels = self.detectors

        return pd.DataFrame(self.matrix, columns=labels, index=labels)



""" transforms module """
from ._transforms import \
    LinearTransform, \
    LogTransform, \
    RatioTransform, \
    HyperlogTransform, \
    LogicleTransform, \
    AsinhTransform
from ._wsp_transforms import WSPLogTransform, WSPBiexTransform

__all__ = [
    'LinearTransform',
    'LogTransform',
    'RatioTransform',
    'HyperlogTransform',
    'LogicleTransform',
    'AsinhTransform',
    'WSPLogTransform',
    'WSPBiexTransform'
]



"""
Utility functions for parsing & exporting gating related XML documents
"""
import numpy as np
from lxml import etree
import networkx as nx
from .._resources import gml_schema
from .._models.dimension import Dimension, RatioDimension, QuadrantDivider
from .._models.vertex import Vertex
from .._models.gating_strategy import GatingStrategy
# noinspection PyProtectedMember
from .._models.transforms import _transforms, _gml_transforms
# noinspection PyProtectedMember
from .._models.transforms._matrix import Matrix
# noinspection PyProtectedMember
from .._models.gates._gml_gates import \
    GMLBooleanGate, \
    GMLEllipsoidGate, \
    GMLQuadrantGate, \
    GMLPolygonGate, \
    GMLRectangleGate
# noinspection PyProtectedMember
from .._models.gates._gates import \
    BooleanGate, \
    EllipsoidGate, \
    QuadrantGate, \
    PolygonGate, \
    RectangleGate


# map GatingML gate keys to our GML gate classes
gate_constructor_lut = {
    'RectangleGate': GMLRectangleGate,
    'PolygonGate': GMLPolygonGate,
    'EllipsoidGate': GMLEllipsoidGate,
    'QuadrantGate': GMLQuadrantGate,
    'BooleanGate': GMLBooleanGate
}


def parse_gating_xml(xml_file_or_path):
    """
    Parse a GatingML 20 document and return as a GatingStrategy.

    :param xml_file_or_path: file handle or file path to a GatingML 2.0 document
    :return: GatingStrategy instance
    """
    doc_type, root_gml, gating_ns, data_type_ns, xform_ns = _get_xml_type(xml_file_or_path)

    gating_strategy = GatingStrategy()

    if doc_type == 'gatingml':
        gates = _construct_gates(root_gml, gating_ns, data_type_ns)
        transformations = _construct_transforms(root_gml, xform_ns, data_type_ns)
        comp_matrices = _construct_matrices(root_gml, xform_ns, data_type_ns)
    elif doc_type == 'flowjo':
        raise ValueError("File is a FlowJo workspace, use parse_wsp or Session.import_flowjo_workspace.")
    else:
        raise ValueError("Gating file format is not supported.")

    for c_id, c in comp_matrices.items():
        gating_strategy.add_comp_matrix(c)
    for t_id, t in transformations.items():
        gating_strategy.add_transform(t)

    deps = []
    quadrants = []
    bool_edges = []

    for g_id, gate in gates.items():
        if gate.parent is None:
            parent = 'root'
        else:
            parent = gate.parent

        deps.append((parent, g_id))

        if isinstance(gate, QuadrantGate):
            for q_id in gate.quadrants:
                deps.append((g_id, q_id))
                quadrants.append(q_id)

        if isinstance(gate, BooleanGate):
            for g_ref in gate.gate_refs:
                deps.append((g_ref['ref'], g_id))

                bool_edges.append((g_ref['ref'], g_id))

    dag = nx.DiGraph()
    dag.add_edges_from(deps)

    is_acyclic = nx.is_directed_acyclic_graph(dag)

    if not is_acyclic:
        raise ValueError("The given GatingML 2.0 file is invalid, cyclic gate dependencies are not allowed.")

    process_order = list(nx.algorithms.topological_sort(dag))

    for q_id in quadrants:
        process_order.remove(q_id)

    # remove boolean edges to create a true ancestor graph
    dag.remove_edges_from(bool_edges)

    for g_id in process_order:
        # skip 'root' node
        if g_id == 'root':
            continue
        gate = gates[g_id]

        # For Boolean gates we need to add gate paths to the
        # referenced gates via 'gate_path' key in the gate_refs dict
        if isinstance(gate, BooleanGate):
            bool_gate_refs = gate.gate_refs
            for gate_ref in bool_gate_refs:
                # since we're parsing GML, all gate IDs must be unique
                # so safe to lookup in our graph
                gate_ref_path = list(nx.all_simple_paths(dag, 'root', gate_ref['ref']))[0]
                gate_ref['path'] = tuple(gate_ref_path[:-1])  # don't repeat the gate name

        gating_strategy.add_gate(gate)

    return gating_strategy


def _get_xml_type(xml_file_or_path):
    xml_document = etree.parse(xml_file_or_path)

    val = gml_schema.validate(xml_document)
    root = xml_document.getroot()

    if val:
        doc_type = 'gatingml'
    else:
        # Try parsing as a FlowJo workspace
        if 'flowJoVersion' in root.attrib:
            if int(root.attrib['flowJoVersion'].split('.')[0]) >= 10:
                doc_type = 'flowjo'
            else:
                raise ValueError("FlowKit only supports FlowJo workspaces for version 10 or higher.")
        else:
            raise ValueError("File is neither GatingML 2.0 compliant nor a FlowJo workspace.")

    gating_ns = None
    data_type_ns = None
    transform_ns = None

    # find GatingML target namespace in the map
    for ns, url in root.nsmap.items():
        if url == 'http://www.isac-net.org/std/Gating-ML/v2.0/gating':
            gating_ns = ns
        elif url == 'http://www.isac-net.org/std/Gating-ML/v2.0/datatypes':
            data_type_ns = ns
        elif url == 'http://www.isac-net.org/std/Gating-ML/v2.0/transformations':
            transform_ns = ns

    if gating_ns is None:
        raise ValueError("GatingML namespace reference is missing from GatingML file")

    return doc_type, root, gating_ns, data_type_ns, transform_ns


def _construct_gates(root_gml, gating_ns, data_type_ns):
    gates_dict = {}

    for gate_str, gate_class in gate_constructor_lut.items():
        gt_gates = root_gml.findall(
            ':'.join([gating_ns, gate_str]),
            root_gml.nsmap
        )

        for gt_gate in gt_gates:
            g = gate_class(
                gt_gate,
                gating_ns,
                data_type_ns
            )

            if g.gate_name in gates_dict:
                raise ValueError(
                    "Gate '%s' already exists. "
                    "Duplicate gate names are not allowed." % g.gate_name
                )
            gates_dict[g.gate_name] = g

    return gates_dict


def _construct_transforms(root_gml, transform_ns, data_type_ns):
    transformations = {}

    if transform_ns is not None:
        # types of transforms include:
        #   - ratio
        #   - log10
        #   - asinh
        #   - hyperlog
        #   - linear
        #   - logicle
        xform_els = root_gml.findall(
            '%s:transformation' % transform_ns,
            namespaces=root_gml.nsmap
        )

        for xform_el in xform_els:
            xform = None

            # determine type of transformation
            fratio_els = xform_el.findall(
                '%s:fratio' % transform_ns,
                namespaces=root_gml.nsmap
            )

            if len(fratio_els) > 0:
                xform = _gml_transforms.RatioGMLTransform(
                    xform_el,
                    transform_ns,
                    data_type_ns
                )

            flog_els = xform_el.findall(
                '%s:flog' % transform_ns,
                namespaces=root_gml.nsmap
            )

            if len(flog_els) > 0:
                xform = _gml_transforms.LogGMLTransform(
                    xform_el,
                    transform_ns
                )

            fasinh_els = xform_el.findall(
                '%s:fasinh' % transform_ns,
                namespaces=root_gml.nsmap
            )

            if len(fasinh_els) > 0:
                xform = _gml_transforms.AsinhGMLTransform(
                    xform_el,
                    transform_ns
                )

            hyperlog_els = xform_el.findall(
                '%s:hyperlog' % transform_ns,
                namespaces=root_gml.nsmap
            )

            if len(hyperlog_els) > 0:
                xform = _gml_transforms.HyperlogGMLTransform(
                    xform_el,
                    transform_ns
                )

            flin_els = xform_el.findall(
                '%s:flin' % transform_ns,
                namespaces=root_gml.nsmap
            )

            if len(flin_els) > 0:
                xform = _gml_transforms.LinearGMLTransform(
                    xform_el,
                    transform_ns
                )

            logicle_els = xform_el.findall(
                '%s:logicle' % transform_ns,
                namespaces=root_gml.nsmap
            )

            if len(logicle_els) > 0:
                xform = _gml_transforms.LogicleGMLTransform(
                    xform_el,
                    transform_ns
                )

            if xform is not None:
                transformations[xform.id] = xform

    return transformations


def _construct_matrices(root_gml, transform_ns, data_type_ns):
    comp_matrices = {}

    if transform_ns is not None:
        # comp matrices are defined by the 'spectrumMatrix' element
        matrix_els = root_gml.findall(
            '%s:spectrumMatrix' % transform_ns,
            namespaces=root_gml.nsmap
        )

        for matrix_el in matrix_els:
            matrix = _parse_matrix_element(
                matrix_el,
                transform_ns,
                data_type_ns
            )

            comp_matrices[matrix.id] = matrix

    return comp_matrices


def find_attribute_value(xml_el, namespace, attribute_name):
    """
    Extract the value from an XML element attribute.

    :param xml_el: lxml etree Element
    :param namespace: string for the XML element's namespace prefix
    :param attribute_name: attribute string to retrieve the value from
    :return: value string for the given attribute_name
    """
    attribs = xml_el.xpath(
        '@%s:%s' % (namespace, attribute_name),
        namespaces=xml_el.nsmap,
        smart_strings=False
    )
    attribs_cnt = len(attribs)

    if attribs_cnt > 1:
        raise ValueError(
            "Multiple %s attributes found (line %d)" % (
                attribute_name, xml_el.sourceline
            )
        )
    elif attribs_cnt == 0:
        return None

    # return as pure str to save memory (otherwise it's an _ElementUnicodeResult from lxml)
    return str(attribs[0])


def parse_gate_element(
        gate_element,
        gating_namespace,
        data_type_namespace
):
    """
    This class parses a GatingML-2.0 compatible gate XML element and extracts the gate ID,
     parent gate ID, and dimensions.

    :param gate_element: gate XML element from a GatingML-2.0 document
    :param gating_namespace: XML namespace for gating elements/attributes
    :param data_type_namespace: XML namespace for data type elements/attributes
    """
    gate_id = find_attribute_value(gate_element, gating_namespace, 'id')
    parent_id = find_attribute_value(gate_element, gating_namespace, 'parent_id')

    # most gates specify dimensions in the 'dimension' tag,
    # but quad gates specify dimensions in the 'divider' tag
    div_els = gate_element.findall(
        '%s:divider' % gating_namespace,
        namespaces=gate_element.nsmap
    )

    dimensions = []  # may actually be a list of dividers

    if len(div_els) == 0:
        dim_els = gate_element.findall(
            '%s:dimension' % gating_namespace,
            namespaces=gate_element.nsmap
        )

        dimensions = []

        for dim_el in dim_els:
            dim = _parse_dimension_element(dim_el, gating_namespace, data_type_namespace)
            dimensions.append(dim)
    else:
        for div_el in div_els:
            dim = _parse_divider_element(div_el, gating_namespace, data_type_namespace)
            dimensions.append(dim)

    return gate_id, parent_id, dimensions


def _parse_dimension_element(
        dim_element,
        gating_namespace,
        data_type_namespace
):
    compensation_ref = find_attribute_value(dim_element, gating_namespace, 'compensation-ref')
    transformation_ref = find_attribute_value(dim_element, gating_namespace, 'transformation-ref')

    range_min = None
    range_max = None

    # should be 0 or only 1 'min' attribute (same for 'max')
    _min = find_attribute_value(dim_element, gating_namespace, 'min')
    _max = find_attribute_value(dim_element, gating_namespace, 'max')

    if _min is not None:
        range_min = float(_min)
    if _max is not None:
        range_max = float(_max)

    # ID be here
    fcs_dim_el = dim_element.find(
        '%s:fcs-dimension' % data_type_namespace,
        namespaces=dim_element.nsmap
    )

    # if no 'fcs-dimension' element is present, this might be a
    # 'new-dimension'  made from a transformation on other dims
    if fcs_dim_el is None:
        new_dim_el = dim_element.find(
            '%s:new-dimension' % data_type_namespace,
            namespaces=dim_element.nsmap
        )
        if new_dim_el is None:
            raise ValueError(
                "Dimension invalid: neither fcs-dimension or new-dimension "
                "tags found (line %d)" % dim_element.sourceline
            )

        # if we get here, there should be a 'transformation-ref' attribute
        ratio_xform_ref = find_attribute_value(new_dim_el, data_type_namespace, 'transformation-ref')

        if ratio_xform_ref is None:
            raise ValueError(
                "New dimensions must provide a transform reference (line %d)" % dim_element.sourceline
            )
        dimension = RatioDimension(
            ratio_xform_ref,
            compensation_ref,
            transformation_ref,
            range_min=range_min,
            range_max=range_max
        )
    else:
        dim_id = find_attribute_value(fcs_dim_el, data_type_namespace, 'name')
        if dim_id is None:
            raise ValueError(
                'Dimension name not found (line %d)' % fcs_dim_el.sourceline
            )

        dimension = Dimension(
            dim_id,
            compensation_ref,
            transformation_ref,
            range_min=range_min,
            range_max=range_max
        )

    return dimension


def _parse_divider_element(divider_element, gating_namespace, data_type_namespace):
    # Get 'id' (present in quad gate dividers)
    divider_id = find_attribute_value(divider_element, gating_namespace, 'id')

    compensation_ref = find_attribute_value(divider_element, gating_namespace, 'compensation-ref')
    transformation_ref = find_attribute_value(divider_element, gating_namespace, 'transformation-ref')

    # ID be here
    fcs_dim_el = divider_element.find(
        '%s:fcs-dimension' % data_type_namespace,
        namespaces=divider_element.nsmap
    )

    dim_id = find_attribute_value(fcs_dim_el, data_type_namespace, 'name')
    if dim_id is None:
        raise ValueError(
            'Divider dimension name not found (line %d)' % fcs_dim_el.sourceline
        )

    values = []  # quad gate dims can have multiple values

    # values in gating namespace, ok if not present
    value_els = divider_element.findall(
        '%s:value' % gating_namespace,
        namespaces=divider_element.nsmap
    )

    for value in value_els:
        values.append(float(value.text))

    divider = QuadrantDivider(divider_id, dim_id, compensation_ref, values, transformation_ref)

    return divider


def parse_vertex_element(vertex_element, gating_namespace, data_type_namespace):
    """
    This class parses a GatingML-2.0 compatible vertex XML element and returns a Vertex object
     parent gate ID, and dimensions.

    :param vertex_element: vertex XML element from a GatingML-2.0 document
    :param gating_namespace: XML namespace for gating elements/attributes
    :param data_type_namespace: XML namespace for data type elements/attributes
    """
    coordinates = []

    coord_els = vertex_element.findall(
        '%s:coordinate' % gating_namespace,
        namespaces=vertex_element.nsmap
    )

    if len(coord_els) != 2:
        raise ValueError(
            'Vertex must contain 2 coordinate values (line %d)' % vertex_element.sourceline
        )

    # should be 0 or only 1 'min' attribute,
    for coord_el in coord_els:
        value = find_attribute_value(coord_el, data_type_namespace, 'value')
        if value is None:
            raise ValueError(
                'Vertex coordinate must have only 1 value (line %d)' % coord_el.sourceline
            )

        coordinates.append(float(value))

    return Vertex(coordinates)


def _parse_matrix_element(
    matrix_element,
    xform_namespace,
    data_type_namespace
):
    matrix_id = find_attribute_value(matrix_element, xform_namespace, 'id')
    fluorochomes = []
    detectors = []
    matrix = []

    fluoro_el = matrix_element.find(
        '%s:fluorochromes' % xform_namespace,
        namespaces=matrix_element.nsmap
    )

    fcs_dim_els = fluoro_el.findall(
        '%s:fcs-dimension' % data_type_namespace,
        namespaces=matrix_element.nsmap
    )

    for dim_el in fcs_dim_els:
        label = find_attribute_value(dim_el, data_type_namespace, 'name')

        if label is None:
            raise ValueError(
                'Dimension name not found (line %d)' % dim_el.sourceline
            )
        fluorochomes.append(label)

    detectors_el = matrix_element.find(
        '%s:detectors' % xform_namespace,
        namespaces=matrix_element.nsmap
    )

    fcs_dim_els = detectors_el.findall(
        '%s:fcs-dimension' % data_type_namespace,
        namespaces=matrix_element.nsmap
    )

    for dim_el in fcs_dim_els:
        label = find_attribute_value(dim_el, data_type_namespace, 'name')

        if label is None:
            raise ValueError(
                'Dimension name not found (line %d)' % dim_el.sourceline
            )
        detectors.append(label)

    spectrum_els = matrix_element.findall(
        '%s:spectrum' % xform_namespace,
        namespaces=matrix_element.nsmap
    )

    for spectrum_el in spectrum_els:
        matrix_row = []

        coefficient_els = spectrum_el.findall(
            '%s:coefficient' % xform_namespace,
            namespaces=matrix_element.nsmap
        )

        for co_el in coefficient_els:
            value = find_attribute_value(co_el, xform_namespace, 'value')
            if value is None:
                raise ValueError(
                    'Matrix coefficient must have only 1 value (line %d)' % co_el.sourceline
                )

            matrix_row.append(float(value))

        matrix.append(matrix_row)

    matrix = np.array(matrix)

    return Matrix(matrix_id, matrix, detectors, fluorochomes)


def _add_matrix_to_gml(root, matrix, ns_map):
    xform_ml = etree.SubElement(root, "{%s}spectrumMatrix" % ns_map['transforms'])
    xform_ml.set('{%s}id' % ns_map['transforms'], matrix.id)

    fluoros_ml = etree.SubElement(xform_ml, "{%s}fluorochromes" % ns_map['transforms'])

    for fluoro in matrix.fluorochomes:
        fluoro_ml = etree.SubElement(fluoros_ml, '{%s}fcs-dimension' % ns_map['data-type'])
        fluoro_ml.set('{%s}name' % ns_map['data-type'], fluoro)

    detectors_ml = etree.SubElement(xform_ml, "{%s}detectors" % ns_map['transforms'])

    for detector in matrix.detectors:
        detector_ml = etree.SubElement(detectors_ml, '{%s}fcs-dimension' % ns_map['data-type'])
        detector_ml.set('{%s}name' % ns_map['data-type'], detector)

    for row in matrix.matrix:
        row_ml = etree.SubElement(xform_ml, "{%s}spectrum" % ns_map['transforms'])
        for val in row:
            coefficient_ml = etree.SubElement(row_ml, "{%s}coefficient" % ns_map['transforms'])
            coefficient_ml.set('{%s}value' % ns_map['transforms'], str(val))


def _add_transform_to_gml(root, transform, ns_map):
    xform_ml = etree.SubElement(root, "{%s}transformation" % ns_map['transforms'])
    xform_ml.set('{%s}id' % ns_map['transforms'], transform.id)

    if isinstance(transform, _transforms.RatioTransform):
        ratio_ml = etree.SubElement(xform_ml, "{%s}fratio" % ns_map['transforms'])
        ratio_ml.set('{%s}A' % ns_map['transforms'], str(transform.param_a))
        ratio_ml.set('{%s}B' % ns_map['transforms'], str(transform.param_b))
        ratio_ml.set('{%s}C' % ns_map['transforms'], str(transform.param_c))

        for dim in transform.dimensions:
            fcs_dim_ml = etree.SubElement(ratio_ml, '{%s}fcs-dimension' % ns_map['data-type'])
            fcs_dim_ml.set('{%s}name' % ns_map['data-type'], dim)
    elif isinstance(transform, _transforms.LogTransform):
        log_ml = etree.SubElement(xform_ml, "{%s}flog" % ns_map['transforms'])
        log_ml.set('{%s}T' % ns_map['transforms'], str(transform.param_t))
        log_ml.set('{%s}M' % ns_map['transforms'], str(transform.param_m))
    elif isinstance(transform, _transforms.AsinhTransform):
        asinh_ml = etree.SubElement(xform_ml, "{%s}fasinh" % ns_map['transforms'])
        asinh_ml.set('{%s}T' % ns_map['transforms'], str(transform.param_t))
        asinh_ml.set('{%s}M' % ns_map['transforms'], str(transform.param_m))
        asinh_ml.set('{%s}A' % ns_map['transforms'], str(transform.param_a))
    elif isinstance(transform, _transforms.LogicleTransform):
        logicle_ml = etree.SubElement(xform_ml, "{%s}logicle" % ns_map['transforms'])
        logicle_ml.set('{%s}T' % ns_map['transforms'], str(transform.param_t))
        logicle_ml.set('{%s}W' % ns_map['transforms'], str(transform.param_w))
        logicle_ml.set('{%s}M' % ns_map['transforms'], str(transform.param_m))
        logicle_ml.set('{%s}A' % ns_map['transforms'], str(transform.param_a))
    elif isinstance(transform, _transforms.HyperlogTransform):
        hyperlog_ml = etree.SubElement(xform_ml, "{%s}hyperlog" % ns_map['transforms'])
        hyperlog_ml.set('{%s}T' % ns_map['transforms'], str(transform.param_t))
        hyperlog_ml.set('{%s}W' % ns_map['transforms'], str(transform.param_w))
        hyperlog_ml.set('{%s}M' % ns_map['transforms'], str(transform.param_m))
        hyperlog_ml.set('{%s}A' % ns_map['transforms'], str(transform.param_a))
    elif isinstance(transform, _transforms.LinearTransform):
        lin_ml = etree.SubElement(xform_ml, "{%s}flin" % ns_map['transforms'])
        lin_ml.set('{%s}T' % ns_map['transforms'], str(transform.param_t))
        lin_ml.set('{%s}A' % ns_map['transforms'], str(transform.param_a))


def _add_gate_to_gml(root, gate, ns_map):
    if isinstance(gate, RectangleGate):
        gate_ml = etree.SubElement(root, "{%s}RectangleGate" % ns_map['gating'])
    elif isinstance(gate, PolygonGate):
        gate_ml = etree.SubElement(root, "{%s}PolygonGate" % ns_map['gating'])

        for v in gate.vertices:
            vert_ml = etree.SubElement(gate_ml, '{%s}vertex' % ns_map['gating'])
            for c in v.coordinates:
                coord_ml = etree.SubElement(vert_ml, '{%s}coordinate' % ns_map['gating'])
                coord_ml.set('{%s}value' % ns_map['data-type'], str(c))
    elif isinstance(gate, BooleanGate):
        gate_ml = etree.SubElement(root, "{%s}BooleanGate" % ns_map['gating'])

        if gate.type == 'and':
            bool_type_ml = etree.SubElement(gate_ml, '{%s}and' % ns_map['gating'])
        elif gate.type == 'or':
            bool_type_ml = etree.SubElement(gate_ml, '{%s}or' % ns_map['gating'])
        elif gate.type == 'not':
            bool_type_ml = etree.SubElement(gate_ml, '{%s}not' % ns_map['gating'])
        else:
            raise ValueError("Boolean gate type '%s' is not valid" % gate.type)

        for gate_ref in gate.gate_refs:
            gate_ref_ml = etree.SubElement(bool_type_ml, '{%s}gateReference' % ns_map['gating'])
            gate_ref_ml.set('{%s}ref' % ns_map['gating'], gate_ref['ref'])
            if gate_ref['complement']:
                gate_ref_ml.set('{%s}use-as-complement' % ns_map['gating'], "true")

    elif isinstance(gate, EllipsoidGate):
        gate_ml = etree.SubElement(root, "{%s}EllipsoidGate" % ns_map['gating'])
        mean_ml = etree.SubElement(gate_ml, '{%s}mean' % ns_map['gating'])
        cov_ml = etree.SubElement(gate_ml, '{%s}covarianceMatrix' % ns_map['gating'])
        dist_square_ml = etree.SubElement(gate_ml, '{%s}distanceSquare' % ns_map['gating'])
        dist_square_ml.set('{%s}value' % ns_map['data-type'], str(gate.distance_square))

        for c in gate.coordinates:
            coord_ml = etree.SubElement(mean_ml, '{%s}coordinate' % ns_map['gating'])
            coord_ml.set('{%s}value' % ns_map['data-type'], str(c))

        for row in gate.covariance_matrix:
            row_ml = etree.SubElement(cov_ml, '{%s}row' % ns_map['gating'])

            for val in row:
                entry_ml = etree.SubElement(row_ml, '{%s}entry' % ns_map['gating'])
                entry_ml.set('{%s}value' % ns_map['data-type'], str(val))
    elif isinstance(gate, QuadrantGate):
        gate_ml = etree.SubElement(root, "{%s}QuadrantGate" % ns_map['gating'])

        for q_id, quadrant in gate.quadrants.items():
            quad_ml = etree.SubElement(gate_ml, '{%s}Quadrant' % ns_map['gating'])
            quad_ml.set('{%s}id' % ns_map['gating'], q_id)

            for div_ref in quadrant.divider_refs:
                pos_ml = etree.SubElement(quad_ml, '{%s}position' % ns_map['gating'])
                pos_ml.set('{%s}divider_ref' % ns_map['gating'], div_ref)

                div_ranges = quadrant.get_divider_range(div_ref)
                if div_ranges[0] is None:
                    loc_coord = div_ranges[1] / 2.0
                elif div_ranges[1] is None:
                    loc_coord = div_ranges[0] * 2.0
                else:
                    loc_coord = np.mean(div_ranges)
                
                pos_ml.set('{%s}location' % ns_map['gating'], str(loc_coord))
    else:
        raise ValueError("Gate %s is not a valid GatingML 2.0 element" % gate.gate_name)

    gate_ml.set('{%s}id' % ns_map['gating'], gate.gate_name)

    for i, dim in enumerate(gate.dimensions):
        dim_type = 'dim'

        if isinstance(dim, QuadrantDivider):
            dim_ml = etree.Element('{%s}divider' % ns_map['gating'])
            dim_ml.set('{%s}id' % ns_map['gating'], dim.id)
            dim_type = 'quad'
        elif isinstance(dim, RatioDimension):
            dim_ml = etree.Element('{%s}dimension' % ns_map['gating'])
            dim_type = 'ratio'
        else:
            dim_ml = etree.Element('{%s}dimension' % ns_map['gating'])

        gate_ml.insert(i, dim_ml)

        if dim.compensation_ref is not None:
            dim_ml.set('{%s}compensation-ref' % ns_map['gating'], dim.compensation_ref)
        if dim.transformation_ref is not None:
            dim_ml.set('{%s}transformation-ref' % ns_map['gating'], dim.transformation_ref)

        if dim_type != 'quad':
            if dim.min is not None:
                dim_ml.set('{%s}min' % ns_map['gating'], str(dim.min))
            if dim.max is not None:
                dim_ml.set('{%s}max' % ns_map['gating'], str(dim.max))

        if dim_type == 'ratio':
            new_dim_el = etree.SubElement(dim_ml, '{%s}new-dimension' % ns_map['data-type'])
            new_dim_el.set('{%s}transformation-ref' % ns_map['data-type'], dim.ratio_ref)
        else:
            fcs_dim_ml = etree.SubElement(dim_ml, '{%s}fcs-dimension' % ns_map['data-type'])
            if dim_type == 'dim':
                fcs_dim_ml.set('{%s}name' % ns_map['data-type'], dim.id)
            elif dim_type == 'quad':
                fcs_dim_ml.set('{%s}name' % ns_map['data-type'], dim.dimension_ref)
                for val in dim.values:
                    value_ml = etree.SubElement(dim_ml, '{%s}value' % ns_map['gating'])
                    value_ml.text = str(val)

    return gate_ml


def _add_gates_from_gate_dict(gating_strategy, gate_dict, ns_map, parent_ml):
    # the gate_dict will have keys 'name' and 'children'. top-level 'name' value is 'root'
    for child in gate_dict['children']:
        gate_id = child['name']
        skip = False

        gate = gating_strategy.get_gate(gate_id)
        if isinstance(gate, QuadrantGate):
            if gate_id in gate.quadrants.keys():
                # single quadrants will be handled in the owning quadrant gate
                skip = True

        if not skip:
            child_ml = _add_gate_to_gml(parent_ml, gate, ns_map)

            if gate_dict['name'] != 'root':
                # this is a recursion, add the parent reference
                child_ml.set('{%s}parent_id' % ns_map['gating'], gate_dict['name'])

        if 'children' in child:  # and not isinstance(gate, QuadrantGate):
            _add_gates_from_gate_dict(gating_strategy, child, ns_map, parent_ml)


def export_gatingml(gating_strategy, file_handle):
    """
    Exports a valid GatingML 2.0 document from given GatingStrategy instance
    :param gating_strategy: A GatingStrategy instance
    :param file_handle: File handle for exported GatingML 2.0 document
    :return: None
    """
    ns_g = "http://www.isac-net.org/std/Gating-ML/v2.0/gating"
    ns_dt = "http://www.isac-net.org/std/Gating-ML/v2.0/datatypes"
    ns_xform = "http://www.isac-net.org/std/Gating-ML/v2.0/transformations"
    ns_map = {
        'gating': ns_g,
        'data-type': ns_dt,
        'transforms': ns_xform
    }

    root = etree.Element('{%s}Gating-ML' % ns_g, nsmap=ns_map)

    # process gating strategy transformations
    for xform_id, xform in gating_strategy.transformations.items():
        _add_transform_to_gml(root, xform, ns_map)

    # process gating strategy compensation matrices
    for matrix_id, matrix in gating_strategy.comp_matrices.items():
        _add_matrix_to_gml(root, matrix, ns_map)

    # get gate hierarchy as a dictionary
    gate_dict = gating_strategy.get_gate_hierarchy('dict')

    # recursively convert all gates to GatingML
    _add_gates_from_gate_dict(gating_strategy, gate_dict, ns_map, root)

    et = etree.ElementTree(root)

    et.write(file_handle, encoding="utf-8", xml_declaration=True, pretty_print=True)



"""
Transform sub-classes for parsing GatingML-2.0 XML gates
These classes are not exposed in the public API, they are
intended for internal use when parsing GatingML-2.0 documents.
"""
from ..._utils import xml_utils
from ._transforms import \
    RatioTransform, \
    LinearTransform, \
    LogTransform, \
    HyperlogTransform, \
    LogicleTransform, \
    AsinhTransform


class RatioGMLTransform(RatioTransform):
    """
    This class parses an XML element from from a GatingML-2.0 document and creates an
    instance compatible with the RatioTransform (see RatioTransform docs for more info).

    :param xform_element: ratio transform XML element from a GatingML-2.0 document
    :param xform_namespace: XML namespace for transform elements/attributes
    :param data_type_namespace: XML namespace for data type elements/attributes
    """
    def __init__(
            self,
            xform_element,
            xform_namespace,
            data_type_namespace
    ):
        t_id = xml_utils.find_attribute_value(xform_element, xform_namespace, 'id')

        f_ratio_els = xform_element.findall(
            '%s:fratio' % xform_namespace,
            namespaces=xform_element.nsmap
        )

        if len(f_ratio_els) == 0:
            raise ValueError(
                "Ratio transform must specify an 'fratio' element (line %d)" % xform_element.sourceline
            )

        # f ratio transform has 3 parameters: A, B, and C
        # these are attributes of the 'fratio' element
        param_a = xml_utils.find_attribute_value(f_ratio_els[0], xform_namespace, 'A')
        param_b = xml_utils.find_attribute_value(f_ratio_els[0], xform_namespace, 'B')
        param_c = xml_utils.find_attribute_value(f_ratio_els[0], xform_namespace, 'C')

        if None in [param_a, param_b, param_c]:
            raise ValueError(
                "Ratio transform must provide an 'A', a 'B', and a 'C' "
                "attribute (line %d)" % f_ratio_els[0].sourceline
            )

        fcs_dim_els = f_ratio_els[0].findall(
            '%s:fcs-dimension' % data_type_namespace,
            namespaces=xform_element.nsmap
        )

        dim_ids = []

        for dim_el in fcs_dim_els:
            dim_id = xml_utils.find_attribute_value(dim_el, data_type_namespace, 'name')

            if dim_id is None:
                raise ValueError(
                    'Dimension name not found (line %d)' % dim_el.sourceline
                )
            dim_ids.append(dim_id)

        RatioTransform.__init__(
            self,
            t_id,
            dim_ids,
            float(param_a),
            float(param_b),
            float(param_c)
        )


class LinearGMLTransform(LinearTransform):
    """
    This class parses an XML element from from a GatingML-2.0 document and creates an
    instance compatible with the LinearTransform (see LinearTransform docs for more info).

    :param xform_element: linear transform XML element from a GatingML-2.0 document
    :param xform_namespace: XML namespace for transform elements/attributes
    """
    def __init__(
            self,
            xform_element,
            xform_namespace
    ):
        t_id = xml_utils.find_attribute_value(xform_element, xform_namespace, 'id')

        f_lin_els = xform_element.findall(
            '%s:flin' % xform_namespace,
            namespaces=xform_element.nsmap
        )

        if len(f_lin_els) == 0:
            raise ValueError(
                "Linear transform must specify an 'flin' element (line %d)" % xform_element.sourceline
            )

        # f linear transform has 2 parameters: T and A
        # these are attributes of the 'flin' element
        param_t = xml_utils.find_attribute_value(f_lin_els[0], xform_namespace, 'T')
        param_a = xml_utils.find_attribute_value(f_lin_els[0], xform_namespace, 'A')

        if None in [param_t, param_a]:
            raise ValueError(
                "Linear transform must provide 'T' and 'A' attributes (line %d)" % f_lin_els[0].sourceline
            )

        LinearTransform.__init__(
            self,
            t_id,
            float(param_t),
            float(param_a)
        )


class LogGMLTransform(LogTransform):
    """
    This class parses an XML element from from a GatingML-2.0 document and creates an
    instance compatible with the LogTransform (see LogTransform docs for more info).

    :param xform_element: log transform XML element from a GatingML-2.0 document
    :param xform_namespace: XML namespace for transform elements/attributes
    """
    def __init__(
            self,
            xform_element,
            xform_namespace
    ):
        t_id = xml_utils.find_attribute_value(xform_element, xform_namespace, 'id')

        f_log_els = xform_element.findall(
            '%s:flog' % xform_namespace,
            namespaces=xform_element.nsmap
        )

        if len(f_log_els) == 0:
            raise ValueError(
                "Log transform must specify an 'flog' element (line %d)" % xform_element.sourceline
            )

        # f log transform has 2 parameters: T and M
        # these are attributes of the 'flog' element
        param_t = xml_utils.find_attribute_value(f_log_els[0], xform_namespace, 'T')
        param_m = xml_utils.find_attribute_value(f_log_els[0], xform_namespace, 'M')

        if None in [param_t, param_m]:
            raise ValueError(
                "Log transform must provide an 'T' attribute (line %d)" % f_log_els[0].sourceline
            )

        LogTransform.__init__(
            self,
            t_id,
            float(param_t),
            float(param_m)
        )


class HyperlogGMLTransform(HyperlogTransform):
    """
    This class parses an XML element from from a GatingML-2.0 document and creates an
    instance compatible with the HyperlogTransform (see HyperlogTransform docs for more info).

    :param xform_element: hyperlog transform XML element from a GatingML-2.0 document
    :param xform_namespace: XML namespace for transform elements/attributes
    """
    def __init__(
            self,
            xform_element,
            xform_namespace
    ):
        t_id = xml_utils.find_attribute_value(xform_element, xform_namespace, 'id')

        hyperlog_els = xform_element.findall(
            '%s:hyperlog' % xform_namespace,
            namespaces=xform_element.nsmap
        )

        if len(hyperlog_els) == 0:
            raise ValueError(
                "Hyperlog transform must specify an 'hyperlog' element (line %d)" % xform_element.sourceline
            )

        # hyperlog transform has 4 parameters: T, W, M, and A
        # these are attributes of the 'hyperlog' element
        param_t = xml_utils.find_attribute_value(hyperlog_els[0], xform_namespace, 'T')
        param_w = xml_utils.find_attribute_value(hyperlog_els[0], xform_namespace, 'W')
        param_m = xml_utils.find_attribute_value(hyperlog_els[0], xform_namespace, 'M')
        param_a = xml_utils.find_attribute_value(hyperlog_els[0], xform_namespace, 'A')

        if None in [param_t, param_w, param_m, param_a]:
            raise ValueError(
                "Hyperlog transform must provide 'T', 'W', 'M', and 'A' "
                "attributes (line %d)" % hyperlog_els[0].sourceline
            )

        HyperlogTransform.__init__(
            self,
            t_id,
            float(param_t),
            float(param_w),
            float(param_m),
            float(param_a)
        )


class LogicleGMLTransform(LogicleTransform):
    """
    This class parses an XML element from from a GatingML-2.0 document and creates an
    instance compatible with the LogicleTransform (see LogicleTransform docs for more info).

    :param xform_element: logicle transform XML element from a GatingML-2.0 document
    :param xform_namespace: XML namespace for transform elements/attributes
    """
    def __init__(
            self,
            xform_element,
            xform_namespace
    ):
        t_id = xml_utils.find_attribute_value(xform_element, xform_namespace, 'id')

        logicle_els = xform_element.findall(
            '%s:logicle' % xform_namespace,
            namespaces=xform_element.nsmap
        )

        if len(logicle_els) == 0:
            raise ValueError(
                "Logicle transform must specify an 'logicle' element (line %d)" % xform_element.sourceline
            )

        # logicle transform has 4 parameters: T, W, M, and A
        # these are attributes of the 'logicle' element
        param_t = xml_utils.find_attribute_value(logicle_els[0], xform_namespace, 'T')
        param_w = xml_utils.find_attribute_value(logicle_els[0], xform_namespace, 'W')
        param_m = xml_utils.find_attribute_value(logicle_els[0], xform_namespace, 'M')
        param_a = xml_utils.find_attribute_value(logicle_els[0], xform_namespace, 'A')

        if None in [param_t, param_w, param_m, param_a]:
            raise ValueError(
                "Logicle transform must provide 'T', 'W', 'M', and 'A' "
                "attributes (line %d)" % logicle_els[0].sourceline
            )

        LogicleTransform.__init__(
            self,
            t_id,
            float(param_t),
            float(param_w),
            float(param_m),
            float(param_a)
        )


class AsinhGMLTransform(AsinhTransform):
    """
    This class parses an XML element from from a GatingML-2.0 document and creates an
    instance compatible with the AsinhTransform (see AsinhTransform docs for more info).

    :param xform_element: asinh transform XML element from a GatingML-2.0 document
    :param xform_namespace: XML namespace for transform elements/attributes
    """
    def __init__(
            self,
            xform_element,
            xform_namespace
    ):
        t_id = xml_utils.find_attribute_value(xform_element, xform_namespace, 'id')

        f_asinh_els = xform_element.findall(
            '%s:fasinh' % xform_namespace,
            namespaces=xform_element.nsmap
        )

        if len(f_asinh_els) == 0:
            raise ValueError(
                "Asinh transform must specify an 'fasinh' element (line %d)" % xform_element.sourceline
            )

        # f asinh transform has 3 parameters: T, M, and A
        # these are attributes of the 'fasinh' element
        param_t = xml_utils.find_attribute_value(f_asinh_els[0], xform_namespace, 'T')
        param_m = xml_utils.find_attribute_value(f_asinh_els[0], xform_namespace, 'M')
        param_a = xml_utils.find_attribute_value(f_asinh_els[0], xform_namespace, 'A')

        if None in [param_t, param_m, param_a]:
            raise ValueError(
                "Asinh transform must provide 'T', 'M', and 'A' attributes (line %d)" % f_asinh_els[0].sourceline
            )

        AsinhTransform.__init__(
            self,
            t_id,
            float(param_t),
            float(param_m),
            float(param_a)
        )
